#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2019-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
This module contains the base class for mystic samplers, and describes
the mystic sampler interface.
"""
__all__  = ['AbstractSampler']

"""
# utility functions
_reset_sampler: reset all
_reset_solved: reset all solved
_sample(reset): reset all/stop/none, then step

# use cases to manage evals/iters
_restart_all_without_step: _reset_sampler
_restart_solved_without_step: _reset_solved
_restart_all_then_step: _sample(reset=True)
_restart_solved_then_step: _sample(reset=None)
_step_without_restart: _sample(reset=False)

# options
_restart_if_solved(solved): (* = 'all' if all else 'solved')
  . all: if all stop, _restart_*_then_step
  . True: if best stop, _restart_*_then_step
  . any: if any stop, _restart_*_then_step 
  . False: if none stop, _restart_*_then_step
  . None: _step_without_restart
"""

class AbstractSampler(object): # derived class per sampling algorithm
    """
AbstractSampler base class for optimizer-directed samplers
    """
    def __init__(self, bounds, model, npts=None, **kwds):
        """
        Inputs:
         bounds -- list[tuples]: (lower,upper) bound for each model input
         model -- function: y = model(x), where x is an input parameter vector
         npts -- int: number of points to sample the model

        NOTE:
          additional keywords (evalmon, stepmon, maxiter, maxfun, dist,
          saveiter, state, termination, constraints, penalty, reducer,
          solver, tightrange, cliprange) are available for use.
          See mystic.ensemble for more details.
        """
        self._bounds = bounds
        self._model = model #FIXME: if None, interpolate
        self._npts = npts

        self._evalmon = kwds.pop('evalmon', None)
        if self._evalmon is None:
            from mystic.monitors import Null
            self._evalmon = Null()
        from collections import defaultdict
        self._kwds = defaultdict(lambda :None)
        self._kwds.update(kwds)

        s = self._init_solver()
        kwd = dict(tight=self._kwds['tightrange'], clip=self._kwds['cliprange'])
        s.SetStrictRanges(*zip(*bounds), **kwd)
        #s.SetObjective(memo) #model) #XXX: ExtraArgs: axis ???
        s.SetObjective(model) #FIXME: ensure cached model

        # apply additional kwds
        solver = self._kwds['solver']
        if solver is not None: s.SetNestedSolver(solver)
        s.SetDistribution(self._kwds['dist'])
        s.SetEvaluationLimits(self._kwds['maxiter'], self._kwds['maxfun'])
        s.SetSaveFrequency(self._kwds['saveiter'], self._kwds['state'])
        termination = self._kwds['termination']
        if termination is not None: s.SetTermination(termination) #XXX:?
        s.SetConstraints(self._kwds['constraints'])
        s.SetPenalty(self._kwds['penalty'])
        s.SetReducer(self._kwds['reducer'], arraylike=True)
        s.SetGenerationMonitor(self._kwds['stepmon']) #XXX: use self._stepmon?

        # pass a copy of the monitor to all instances
        import copy
        m = copy.deepcopy(self._evalmon) #XXX: no change with direct reference
        s.SetEvaluationMonitor(m)
        self._sampler = s
        self._npts = s._npts

        self._evals = [0] * s._npts #XXX: collect count or monitor?
        self._iters = [0] * s._npts #XXX: collect count or monitor?
        return


    def _init_solver(self):
        """initialize the ensemble solver"""
        from mystic.abstract_ensemble_solver import AbstractEnsembleSolver
        return AbstractEnsembleSolver(len(self._bounds), npts=self._npts)


    def _reset_sampler(self):
        """reinitialize the ensemble solver/sampler"""
        s = self._sampler
        # the following "seems" to work; all/more necessary?
        #XXX: add this method as internal method for ensemble solver?
        s.popEnergy = [s._init_popEnergy] * s.nPop
        s.population = [[0.0 for i in range(s.nDim)] for j in range(s.nPop)]
        # s.trialSolution = [0.0] * s.nDim
        s._bestEnergy = None
        s._bestSolution = None
        s._fcalls = [0] #XXX:?
        s._allSolvers = [None]*s._npts
        s._bestSolver = None
        s.SetGenerationMonitor(s._stepmon[:0], new=True) #XXX:?
        s.SetEvaluationMonitor(s._evalmon[:0], new=True) #XXX:?
        s._live = False
        s._bootstrap_objective()
        return


    def _reset_solved(self):
        """reinitialize all terminated solvers in the ensemble"""
        s = self._sampler
        # the following "seems" to work; all/more necessary?
        #XXX: add this method as internal method for ensemble solver?
        from copy import copy as _copy
        solver = s._AbstractEnsembleSolver__get_solver_instance(reset=True)
        #solver.SetEvaluationMonitor(solver._evalmon[:0], new=True)
        #solver.SetGenerationMonitor(solver._stepmon[:0], new=True)
        [s._allSolvers.__setitem__(i, _copy(solver)) for i,j in enumerate(s.Terminated(all=True)) if j is True]
        s._bestSolver = None
        s._AbstractEnsembleSolver__update_state()
        return


    def _sample(self, reset=False):#, **kwds):
        """collect a sample for each member in the ensemble

        Inputs:
         reset: if True, reset all before sampling; if None, reset terminated
        """
        #NOTE: sample(reset_all) is like _sample(reset), with None <=> False
        s = self._sampler
        m = self._evalmon

        _eval = s._all_evals
        _iter = s._all_iters
        if reset: self._reset_sampler()
        elif reset is None: self._reset_solved()

        # get evals before taking a step (needed for backfill)
        evals = [len(getattr(i, '_evalmon', ())) for i in s._allSolvers]
        neval = len(m)

        # take a step
        s.Step()
        termination = None #s.Terminated(**kwds) #XXX:?
        #NOTE: i._evalmon._id for i in _allSolvers is [None]

        # backfill original monitor instance #NOTE: not in chronological order
        map(m.extend, [i._evalmon[e:] for e,i in zip(evals, s._allSolvers)])

        _eval = (max(j-i,1) if j>=i else i for (i,j) in zip(_eval,s._all_evals))
        _iter = (max(j-i,1) if j>=i else i for (i,j) in zip(_iter,s._all_iters))
        for i,(e,t) in enumerate(zip(_eval,_iter)):
            self._evals[i] += e
            self._iters[i] += t
        return termination


    def sample(self, if_terminated=None, reset_all=True):#, **kwds):
        """sample npts using vectorized solvers

        Input:
         if_terminated: the amount of termination [all,True,any,False,None]
         reset_all: action to take when if_terminated is met [True,False,None]

        Note:
         When if_terminated is all, reset if all solvers are terminated.
         When if_terminated is any, reset if any solvers are terminated.
         When if_terminated is True, reset if the best solver is terminated.
         When if_terminated is False, reset if no solvers are terminated.
         When if_terminated is None, reset regardless of termination [default].

        Note:
         If reset_all is True, reset all solvers if if_terminated is satisfied.
         If reset_all is False, similarly reset only the terminated solvers.
         If reset_all is None, never reset.
        """
        if type(reset_all) is bool:
            s = self._sampler
            if if_terminated is None: # reset regardless
                return self._sample(reset=(True if reset_all else None))#, **kwds)
            if not if_terminated in (all, True, any, False):
                msg = "%s not in (all, True, any, False, None)" % if_terminated
                raise ValueError(msg)
            stop = s.Terminated(all=True)
            if all(stop) \
            or (stop[s._is_best() or 0] and if_terminated is not all) \
            or (any(stop) and if_terminated in (any, False)) \
            or (not any(stop) and if_terminated is False):
                return self._sample(reset=(True if reset_all else None))#, **kwds)
            # othewise just take a step
            return self._sample(reset=False)#, **kwds)
        return self._sample(reset=False)#, **kwds)


    def sample_until(self, iters=None, evals=None, terminated=None, **kwds):
        """sample until one of the three given conditions is met:
        (iters() >= iters) or (evals() >= evals) or (# terminated > terminated)

        Input:
         iters -- int: maximum number of iterations [default = inf]
         evals -- int: maximum number of evaluations [default = inf]
         terminated -- int: maximum number of stopped solvers [default = inf]

        Additional Input:
         if_terminated: the amount of termination [all,True,any,False,None]
         reset_all: action to take when if_terminated is met [True,False,None]

        Note:
         The default sampler configuration is to always reset (reset_all=True).
         If termination != None, the default is never reset (reset_all=None).
         A limit has to be provided for either iters, evals, or termination.

        Note:
         terminated - {all: all, True: best, any: 1, False: 0, None: inf, N: N}
         if_terminated - {all: all, True: best, any: 1, False: 0, None: always}
         reset_all - {True: reset all, False: reset solved, None: never reset}
        """
        from numpy import inf, all as all_, any as any_
        if evals is None:
            evals = inf
        if iters is None:
            iters = inf
        if terminated is None:
            terminated = self._npts + 1
        else: #NOTE: if termination is specified, don't reset by default
            kwds['reset_all'] = kwds.get('reset_all', None)
            if terminated is all or terminated is all_:
                terminated = self._npts
            elif terminated is any or terminated is any_:
                terminated = 1
            elif terminated is True: # best
                pass
            elif terminated is False:
                terminated = 0
            elif terminated > self._npts:
                msg = 'terminated = %s is greater than npts = %s' % (terminated, self._npts)
                raise ValueError(msg)
        if_stop = kwds.pop('if_terminated', None)
        quit = inf
        if if_stop is all or if_stop is all_:
            quit = self._npts
        elif if_stop is any or if_stop is any_:
            quit = 1
        elif if_stop is True: # best
            quit = True
        elif if_stop is False:
            quit = 0
        elif if_stop is None:
            quit = 0 #XXX: -1 ?
        reset = kwds.pop('reset_all', True)
        # require iters or evals if: (term > npts) OR (reset!=None & quit<term)
        if iters is inf and evals is inf and (terminated > self._npts or
            (reset is not None and quit < terminated)):
            msg = 'either iters = %s or evals = %s must be less than inf' % (iters,evals)
            raise ValueError(msg)
        ###stop = self.terminated(**kwds)
        while self.iters() < iters and self.evals() < evals and (not self.terminated(all=False) if (terminated is True) else (sum(self.terminated(all=True)) < terminated)): #FIXME: sometimes limiting iters stops very late
            stop = self.sample(if_stop, reset)#, **kwds)
        return### stop


    def evals(self, all=False): #XXX: None?
        """get the number of function evaluations

        Input:
          all -- bool: if True, get evals from the entire ensemble
        """
        if all:
            return self._evals
        return sum(self._evals)


    def iters(self, all=False): #XXX: None?
        """get the number of solver iterations

        Input:
          all -- bool: if True, get iters from the entire ensemble
        """
        if all:
            return self._iters
        return sum(self._iters)


    def terminated(self, *args, **kwds): #FIXME: confusing wrt if_terminated?
        """
        Input:
         all, disp, info

        Note:
         all - {True: show all, False: best terminated, None: all terminated}
         disp - {True: show details, False: show less, 'verbose': show more}
         info - {True: return info string, False: return bool}
        """
        _all = kwds.get('all', None)
        if _all not in (True, False, None):
            msg = 'all = %s, must be one of (True, False, None)' % getattr(_all, '__name__', _all)
            raise ValueError(msg)
        kwds['all'] = _all
        return self._sampler.Terminated(*args, **kwds)





#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from g13 import objective, bounds, xs, _xs, x_s, xs_, ys

from mystic.constraints import as_constraint
from mystic.penalty import quadratic_equality

def penalty1(x): # = 0.0
    return x[0]**2 + x[1]**2 + x[2]**2 + x[3]**2 + x[4]**2 - 10.0

def penalty2(x): # = 0.0
    return x[1]*x[2] - 5.0*x[3]*x[4]

def penalty3(x): # = 0.0
    return x[0]**3 + x[1]**3 + 1.0

@quadratic_equality(penalty1)
@quadratic_equality(penalty2)
@quadratic_equality(penalty3)
def penalty(x):
    return 0.0

solver = as_constraint(penalty)



if __name__ == '__main__':

    from mystic.solvers import lattice
    from mystic.math import almostEqual

    result = lattice(objective, 5, [2]*5, bounds=bounds, penalty=penalty, ftol=1e-8, xtol=1e-8, disp=False, full_output=True)

    assert almostEqual(result[0], xs, tol=5e-2) \
        or almostEqual(result[0], _xs, tol=5e-2) \
        or almostEqual(result[0], x_s, tol=5e-2) \
        or almostEqual(result[0], xs_, tol=5e-2)
    assert almostEqual(result[1], ys, rel=5e-2)

# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Testing the polynomial fitting problem of [1] using scipy's Nelder-Mead algorithm.

Reference:

[1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
Heuristic for Global Optimization over Continuous Spaces. Journal of Global
Optimization 11: 341-359, 1997.
"""

from test_ffit import Chebyshev8, ChebyshevCost, plot_solution, print_solution

if __name__ == '__main__':
    import random
    from mystic.solvers import fmin
   #from mystic._scipyoptimize import fmin
    from mystic.tools import random_seed
    random_seed(123)
    x = [random.uniform(-100,100) + Chebyshev8[i] for i in range(9)]
    solution = fmin(ChebyshevCost, x)
    print_solution(solution)
    plot_solution(solution)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2019-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
finds npts that are rtol dist away from legacy data
"""
#XXX: could make similar that uses partitioning (based on monitor.y) ?
#XXX: could make similar that first samples, then uses hole-filling ?

#data=[[2,-6],[5,-7],[1,0],[1,3],[5,9],[2,2],[-10,-10],[-10,10],[10,-10]]
#data=[[0,0],[-10,-10],[-10,0],[10,0],[0,-10],[0,10],[10,-10],[-10,10],[10,10]]
data=[[0,0]]
#data = []

ndim = len(data[0]) if data else 2
lb,ub = [-10]*ndim, [10]*ndim

npts = 12
rtol = None # 4.5

#from mystic.math import Distribution
#from scipy.stats import norm
dist = None # Distribution(norm,0,2)


if __name__ == '__main__':
    from mystic.math import fillpts
    pts = fillpts(lb, ub, npts, data, rtol, dist)

    import numpy as np
    print(np.array(pts).reshape(-1,ndim))

    import matplotlib.pyplot as plt

    _x,_y = np.array(pts).reshape(-1,ndim).T
    x,y = np.array(data).reshape(-1,ndim).T

    plt.scatter(x=x,y=y)
    plt.scatter(x=_x,y=_y)
    plt.show()




#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    x0,x1 = x
    from math import sin, pi
    return -(sin(2*pi*x0)**3 * sin(2*pi*x1)) / (x0**3 * (x0 + x1))

bounds = [(0,10)]*2
# with penalty='penalty' applied, solution is:
xs = [1.22797136, 4.24537337]
ys = -0.09582504

from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.symbolic import generate_penalty, generate_conditions

equations = """
x0**2 - x1 + 1.0 <= 0.0
1.0 - x0 + (x1 - 4)**2 <= 0.0
"""
cf = generate_constraint(generate_solvers(simplify(equations)))
pf = generate_penalty(generate_conditions(equations), k=1e12)



if __name__ == '__main__':

    from mystic.solvers import buckshot
    from mystic.math import almostEqual

    result = buckshot(objective, 2, 40, bounds=bounds, penalty=pf, disp=False, full_output=True)

    assert almostEqual(result[0], xs, tol=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
example of using DifferentialEvolutionSolver on the rosenbrock function
"""

from mystic.models import rosen
import numpy

def constrain(x):
  x[1] = x[0]
  return x

if __name__=='__main__':
    import time

    times = []
    algor = []
    x0 = [0.8,1.2,0.7]
   #x0 = [0.8,1.2,1.7]                  #... better when using "bad" range
    min = [-0.999, -0.999, 0.999]       #XXX: behaves badly when large range
    max = [200.001, 100.001, numpy.inf] #... for >=1 x0 out of bounds; (up xtol)
  # min = [-0.999, -0.999, -0.999]
  # max = [200.001, 100.001, numpy.inf]
 #  min = [-0.999, -0.999, 0.999]
 #  max = [2.001, 1.001, 1.001]
    npop = 5*len(x0)
    print("Differential Evolution")
    print("======================")
    start = time.time()
    from mystic.monitors import Monitor, VerboseMonitor
    stepmon = VerboseMonitor(1,1)
   #stepmon = Monitor() #VerboseMonitor(10)
    from mystic.termination import NormalizedChangeOverGeneration as NCOG

   #from mystic.solvers import diffev, DifferentialEvolutionSolver
    from mystic.solvers import diffev2, DifferentialEvolutionSolver2
   #print(diffev2(rosen,x0,npop,retall=0,full_output=0)#,maxiter=14))
    solver = DifferentialEvolutionSolver2(len(x0), npop)
    solver.SetInitialPoints(x0)
    solver.SetStrictRanges(min,max)
   #solver.SetEvaluationLimits(generations=13)
    solver.SetGenerationMonitor(stepmon)
    solver.SetConstraints(constrain)
    solver.enable_signal_handler()
    solver.Solve(rosen, NCOG(tolerance=1e-4), disp=1)
    print(solver.bestSolution)
   #print("Current function value: %s" % solver.bestEnergy)
   #print("Iterations: %s" % solver.generations)
   #print("Function evaluations: %s" % solver.evaluations)

    times.append(time.time() - start)
    algor.append("Differential Evolution\t")

    for k,t in zip(algor,times):
        print("%s\t -- took %s" % (k, t))

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example use of Forward Mogi Model
in mystic and PARK optimization frameworks.
(built for mystic "trunk" and with park-1.2)

for help, type "python rosetta_mogi_example.py --help"
"""

from math import pi
from numpy import array

try: # check if park is installed
    import park
   #import park.parksnob
    import park.parkde
    Model = park.Model
    __park = True
except ImportError:
    Model = object
    __park = False


def ForwardMogiFactory(params):
    x0,y0,z0,dV = params
    def forward_mogi(evalpts):
        """ evalpts should be a 2D (2 by N) numpy array """
        dx = evalpts[0,:] - x0
        dy = evalpts[1,:] - y0
        dz = 0 - z0
        c = dV * 3. / 4. * pi
        # or equivalently c= (3/4) a^3 dP / rigidity
        # where a = sphere radius, dP = delta Pressure
        r2 = dx*dx + dy*dy + dz*dz
        C = c / pow(r2, 1.5)
        return array((C*dx,C*dy,C*dz))
    return forward_mogi


# --- Cost Function stuff ---
def filter_for_zdisp(input):
    return -input[2,:]

# Here is the cost function
def vec_cost_function(params):
    model = ForwardMogiFactory(params)
    zdisp = filter_for_zdisp(model(stations))
    return 100. * (zdisp - data_z)

# Here is the normed version  [NOTE: fit this one!]
def cost_function(params):
    x = vec_cost_function(params)
    return numpy.sum(real((conjugate(x)*x)))

# a cost function with parameters "normalized"
def vec_cost_function2(params):
    sca = numpy.array([1000, 100., 10., 0.1])
    return vec_cost_function(sca * params)
# --- Cost Function end ---


# --- Plotting stuff ---
import matplotlib.pyplot as plt
def plot_sol(params,linestyle='b-'):
    forward_solution = ForwardMogiFactory(params)
    xx = arange(-30,30,0.1)+actual_params[0]
    yy = 0*xx + actual_params[1]
    ss  = array((xx, yy))
    dd = forward_solution(ss)
    plt.plot(ss[0,:],-dd[2,:],'%s'%linestyle,linewidth=2.0)

def plot_noisy_data():
    import matplotlib.pyplot as plt
    plt.plot(stations[0,:],-data[2,:]+noise[2,:],'k.')
# --- Plotting end ---


# --- Call to Mystic's Fmin optimizer ---
def mystic_optimize(point):
    from mystic.monitors import Monitor, VerboseMonitor
    from mystic.tools import getch, random_seed
    random_seed(123)
    from mystic.solvers import NelderMeadSimplexSolver as fmin
    from mystic.termination import CandidateRelativeTolerance as CRT
    simplex, esow = VerboseMonitor(50), Monitor()
    solver = fmin(len(point))
    solver.SetInitialPoints(point)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(simplex)
    solver.Solve(cost_function, CRT())
    solution = solver.Solution()
    return solution
# --- Mystic end ---

# --- Call to Mystic's DE optimizer ---
def mystic_optimize2(point):
    from mystic.monitors import Monitor, VerboseMonitor
    from mystic.tools import getch, random_seed
    random_seed(123)
    from mystic.solvers import DifferentialEvolutionSolver as de
    from mystic.termination import ChangeOverGeneration as COG
    NPOP = 50
    simplex, esow = VerboseMonitor(50), Monitor()
    solver = de(len(point),NPOP)
    solver.SetInitialPoints(point)
    solver.SetEvaluationMonitor(esow)
    solver.SetGenerationMonitor(simplex)
    solver.Solve(cost_function, COG(generations=100), \
                 CrossProbability=0.5, ScalingFactor=0.5)
    solution = solver.Solution()
    return solution
# --- Mystic end ---

# --- Call to Park ---
class MogiModel(Model):
    """a park model:
 - parameters are passed as named strings to set them as class attributes
 - function that does the evaluation must be named "eval"
 - __call__ generated that takes namestring and parameter-named keywords
"""
    parameters = ["x0","y0","z0","dV"]
    def eval(self, x):
        x0 = self.x0
        y0 = self.y0
        z0 = self.z0
        dV = self.dV
        f = ForwardMogiFactory((x0,y0,z0,dV))
        return f(x)
    pass

class Data2D(object):
    """2d model data with the required park functions"""
    def __init__(self,z):
        self.z = z
        return

    def residuals(self,model):
        zdisp = filter_for_zdisp(model(stations))
        return (100. * (zdisp - self.z)).flatten()
    pass


def park_optimize(point):
    # build the data instance
    data2d = Data2D(data_z)

    # build the model instance
    x0,y0,z0,dV = point 
    model = MogiModel("mymodel",x0=x0,y0=y0,z0=z0,dV=dV)
    # required to set bounds on the parameters
   #model.x0 = [-numpy.inf,numpy.inf]
   #model.y0 = [-numpy.inf,numpy.inf]
   #model.z0 = [-numpy.inf,numpy.inf]
   #model.dV = [-numpy.inf,numpy.inf]
    model.x0 = [-5000,5000]
    model.y0 = [-5000,5000]
    model.z0 = [-5000,5000]
    model.dV = [-5000,5000]

    # add a monitor, and set to print results to the console
    handler=park.fitresult.ConsoleUpdate()

    # select the fitter, and do the fit
   #fitter=park.parksnob.Snobfit()
    fitter=park.parkde.DiffEv()
    # 'fit' requires a list of tuples of (model,data)
    result=park.fit.fit([(model,data2d)],fitter=fitter,handler=handler)

    # print results
   #print(result.calls)    # print number of function calls
   #result.print_summary() # print solution

    # get the results back into a python object
    solution = {}
    for fitparam in result.parameters:
        solution[fitparam.name] = fitparam.value
    solution = [ solution['mymodel.x0'],
                 solution['mymodel.y0'],
                 solution['mymodel.z0'],
                 solution['mymodel.dV'] ]
    return solution
# --- Park end ---


if __name__ == '__main__':
    # parse user selection to solve with "mystic" [default] or "park"
    # also can select mystic's optimizer: "diffev" or "fmin" [default]
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-p","--park",action="store_true",dest="park",\
                      default=False,help="solve with park (instead of mystic)")
    parser.add_option("-m","--mystic",action="store_true",dest="mystic",\
                      default=False,help="solve with mystic's DE optimizer)")
    parsed_opts, parsed_args = parser.parse_args()

    from numpy import pi, sqrt, array, mgrid, random, real, conjugate, arange
    from numpy.random import rand
    import numpy
    # Let the "actual parameters" be :
    actual_params = [1234.,-500., 10., .1]
    actual_forward = ForwardMogiFactory(actual_params)
    print("Target: %s" % actual_params)

    # The data to be "fitted" 
    xstations = array([random.uniform(-30,30) for i in range(300)])+actual_params[0]
    ystations =  0*xstations + actual_params[1]
    stations  = array((xstations, ystations))
    data = actual_forward(stations)

    # generate noise... gaussian distribution with mean 0, sig 0.1e-3
    noise =  array([[random.normal(0,0.1e-3) for i in range(data.shape[1])] for j in range(data.shape[0])])

    # Here is the "observed data"
    data_z = -data[2,:] + noise[2,:]

    # plot the noisy data
    plot_noisy_data()

    point = [1000,-100,0,1] # cg will do badly on this one

    # DO OPTIMIZATION STUFF HERE TO GET SOLUTION
    if parsed_opts.park:     #solve with park's DE
        if __park:
            print("Solving with park's DE optimizer...")
            solution = park_optimize(point)
        else:
            print('This option requires park to be installed')
            exit()
    elif parsed_opts.mystic: #solve with mystic's DE
        print("Solving with mystic's DE optimizer...")
        solution = mystic_optimize2(point)
    else:                    #solve with mystic's fmin
        print("Solving with mystic's fmin optimizer...")
        solution = mystic_optimize(point)
    print("Solved: %s" % solution)

    # plot the solution
    plot_sol(solution,'r-')
    plt.show()


# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
combine several penalty conditions to build a single constraint solver,
then show examples of using (and_, or_, not_) for penalties and constraints
'''

from mystic.coupler import and_, or_, not_
from mystic.constraints import and_ as _and, or_ as _or, not_ as _not
 

if __name__ == '__main__':
    import numpy as np
    from mystic.penalty import linear_equality, quadratic_equality
    from mystic.constraints import as_constraint

    x = x1,x2,x3 = (5., 5., 1.)
    f = f1,f2,f3 = (np.sum, np.prod, np.average)

    k = 100
    solver = 'fmin_powell' #'diffev'
    ptype = quadratic_equality



    # case #1: couple penalties into a single constraint

#   p = [lambda x: abs(xi - fi(x)) for (xi,fi) in zip(x,f)] #XXX
    p1 = lambda x: abs(x1 - f1(x))
    p2 = lambda x: abs(x2 - f2(x))
    p3 = lambda x: abs(x3 - f3(x))
    p = (p1,p2,p3)
    p = [ptype(pi)(lambda x:0.) for pi in p]
    penalty = and_(*p, k=k)
    constraint = as_constraint(penalty, solver=solver)

    x = [1,2,3,4,5]
    x_ = constraint(x)

#   print("target: %s, %s, %s" % (x1, x2, x3))
#   print("solved: %s, %s, %s" % (f1(x_), f2(x_), f3(x_)))
    assert round(f1(x_)) == round(x1)
    assert round(f2(x_)) == round(x2)
    assert round(f3(x_)) == round(x3)



   # case #2: couple constraints into a single constraint

    from mystic.math.measures import impose_product, impose_sum, impose_mean
    from mystic.constraints import as_penalty
    from mystic import random_seed
    random_seed(123)

    t = t1,t2,t3 = (impose_sum, impose_product, impose_mean)
   #c = [lambda x: ti(xi, x) for (xi,ti) in zip(x,t)] #XXX
    c1 = lambda x: t1(x1, x)
    c2 = lambda x: t2(x2, x)
    c3 = lambda x: t3(x3, x)
    c = (c1,c2,c3)
    
    k=1
    solver = 'buckshot' #'diffev'
    ptype = linear_equality #quadratic_equality

    p = [as_penalty(ci, ptype) for ci in c]
    penalty = and_(*p, k=k)
    constraint = as_constraint(penalty, solver=solver)

    x = [1,2,3,4,5]
    x_ = constraint(x)

#   print("target: %s, %s, %s" % (x1, x2, x3))
#   print("solved: %s, %s, %s" % (f1(x_), f2(x_), f3(x_)))
    assert round(f1(x_)) == round(x1)
    assert round(f2(x_)) == round(x2)
    assert round(f3(x_)) == round(x3)



    # etc: more coupling of constraints
    from mystic.constraints import with_mean, discrete

    @with_mean(5.0)
    def meanie(x):
      return x
    
    @discrete(list(range(11)))
    def integers(x):
      return x
    
    c = _and(integers, meanie)
    x = c([1,2,3])
    assert x == integers(x) == meanie(x)
    x = c([9,2,3])
    assert x == integers(x) == meanie(x)
    x = c([0,-2,3])
    assert x == integers(x) == meanie(x)
    x = c([9,-200,344])
    assert x == integers(x) == meanie(x)

    c = _or(meanie, integers)
    x = c([1.1234, 4.23412, -9])
    assert x == meanie(x) and x != integers(x)
    x = c([7.0, 10.0, 0.0])
    assert x == integers(x) and x != meanie(x)
    x = c([6.0, 9.0, 0.0])
    assert x == integers(x) == meanie(x)
    x = c([3,4,5])
    assert x == integers(x) and x != meanie(x)
    x = c([3,4,5.5])
    assert x == meanie(x) and x != integers(x)

    c = _not(integers)
    x = c([1,2,3])
    assert x != integers(x) and x != [1,2,3] and x == c(x)
    x = c([1.1,2,3])
    assert x != integers(x) and x == [1.1,2,3] and x == c(x)
    c = _not(meanie)
    x = c([1,2,3])
    assert x != meanie(x) and x == [1,2,3] and x == c(x)
    x = c([4,5,6])
    assert x != meanie(x) and x != [4,5,6] and x == c(x)
    c = _not(_and(meanie, integers))
    x = c([4,5,6])
    assert x != meanie(x) and x != integers(x) and x != [4,5,6] and x == c(x)


    # etc: more coupling of penalties
    from mystic.penalty import quadratic_inequality

    p1 = lambda x: sum(x) - 5
    p2 = lambda x: min(i**2 for i in x)
    p = p1,p2

    p = [quadratic_inequality(pi)(lambda x:0.) for pi in p]
    p1,p2 = p
    penalty = and_(*p)

    x = [[1,2],[-2,-1],[5,-5]]
    for xi in x:
        assert p1(xi) + p2(xi) == penalty(xi)

    penalty = or_(*p)
    for xi in x:
        assert min(p1(xi),p2(xi)) == penalty(xi)

    penalty = not_(p1)
    for xi in x:
        assert bool(p1(xi)) != bool(penalty(xi))
    penalty = not_(p2)
    for xi in x:
        assert bool(p2(xi)) != bool(penalty(xi))




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = _doc = """
This is drawn from examples in the NAG Library, with the 'peaks' function
definition found in [1].

References:
    1. Numerical Algorithms Group, "NAG Library", Oxford UK, Mark 24,
       2013. http://www.nag.co.uk/numeric/CL/nagdoc_cl24/pdf/E05/e05jbc.pdf
"""
from .abstract_model import AbstractFunction

from math import exp

class Peaks(AbstractFunction):
    __doc__ = \
    """a peaks function generator

A peaks function [1] is essentially flat, with three wells and three
peaks near the origin. The global minimum is separated from the local
minima by peaks.

The generated function f(x) is identical to the 'peaks' function in
section 10 of [1], and requires len(x) == 2.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates an 2-dimensional peaks function for a list of coeffs

f(x) = f_0(x) - f_1(x) - f_2(x)

Where:
f_0(x) = 3 * (1 - x_0)^2 * \exp(-x_0^2 - (x_1 + 1)^2)
and
f_1(x) = 10 * (.2 * x_0 - x_0^3 - x_1^5) * \exp(-x_0^2 - x_1^2)
and
f_2(x) = \exp(-(x_0 + 1)^2 - x_1^2) / 3

Inspect with mystic_model_plotter using::
    mystic.models.peaks -b "-5:5:.1, -5:5:.1" -d

The minimum is f(x)=-6.551133332835841 at x=(0.22827892, -1.62553496)"""
       #x = asarray(x) #XXX: converting to numpy.array slows by 10x
        x,y = coeffs
        result = 3.*(1. - x)**2*exp(-x**2 - (y + 1.)**2) - \
                10.*(x*(1./5.) - x**3 - y**5)*exp(-x**2 - y**2) - \
                1./3.*exp(-(x + 1.)**2 - y**2)
        return result

    minimizers = [(0.22827892, -1.62553496),(-1.34739625,  0.20451886),(0.29644556,  0.3201962)]
    pass

# cleanup
del _doc

# prepared instances
peaks = Peaks().function

# End of file



#!/usr/bin/env python
#
# Problem definition and original response:
# https://stackoverflow.com/a/71197490/2379433
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Minimize y = f(x):
  where y > g(x)

  with f(x) = x0^(sin(x0)) + x1^4 + 6x1^3 - 5x1^2 - 40x1 + 35
  and g(x) = 7x1 - x0 + 5
  in x = [0,10]
"""
import numpy as np
import mystic as my


def f(x):
    x0,x1 = x
    return x0**np.sin(x0) + x1**4 + 6*x1**3 - 5*x1**2 - 40*x1 + 35

def g(x):
    x0,x1 = x
    return 7*x1 - x0 + 5

def penalty(x):  # <= 0.0
    return g(x) - f(x)

@my.penalty.quadratic_inequality(penalty, k=1e12)
def p(x):
    return 0.0

# monitor
mon = my.monitors.VerboseMonitor(1,1)
my.solvers.fmin(f, [5,5], bounds=[(0,10)]*2, penalty=p, itermon=mon, disp=1)

my.log_reader(mon)

fig = my.model_plotter(f, mon, depth=True, scale=1.0, bounds="0:10, 0:10", out=True)

from matplotlib import cm
import matplotlib.pyplot as plt

x,y = my.scripts._parse_axes("0:10, 0:10", grid=True)
x, y = np.meshgrid(x, y)
z = 0*x
s,t = x.shape
for i in range(s):
    for j in range(t):
        xx,yy = x[i,j], y[i,j]
        z[i,j] = g([xx,yy])

z = np.log(4*z*1.0+1)+2 # scale=1.0

ax = fig.axes[0]
ax.contour(x, y, z, 50, cmap=cm.cool)
plt.show()



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import mystic.symbolic as ms


if __name__ == '__main__':

  eqn = '''
  abs(1 - x0) - abs(1 + abs(x0) - abs(x1 + x2)) > 4
  x0**.5 = abs(x1 + x2)
  abs(x0) - x1 > 0
  x2**2 > 1
  (x2**2 - 1)**.5 + x1 > 0
  '''

  res = ms.absval(eqn, all=True)
  assert len(res) == 64
  assert not any('abs(' in r for r in res)
  #print(res[0])
  #print('')

  eq_ = '''
  abs(x0) - x1 > 0
  x2/(x0 - x1) > 1
  '''

  res = ms.simplify(eq_, all=True)
  #print(res[0])
  #print('')
  res0 = ms.simplify(eq_, all=False)
  #print(res0)
  #print('')
  assert res0 in res
  assert len(res) == 3

  _eq = '''
  x0**.5 == abs(x1 + x2)
  abs(x0) - x1 > 0
  x2**2 > 1
  '''

  res = ms.simplify(_eq, all=True)
  import mystic.constraints as mc
  c = ms.generate_constraint(ms.generate_solvers(res), join=mc.or_)
  kx = c([1,2,3])
  #print(kx)
  x0,x1,x2 = kx
  assert x0**.5 == abs(x1 + x2)
  assert abs(x0) - x1 > 0
  assert x2**2 > 1




#!/usr/bin/env python
#
# Problem definition:
# Example in google/or-tools
# https://github.com/google/or-tools/blob/master/examples/python/crypto.py
# with Copyright 2010 Hakan Kjellerstrand hakank@bonetmail.com
# and disclamer as stated at the above reference link.
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
  Crypto problem in Google CP Solver.

  Prolog benchmark problem
  '''
  Name           : crypto.pl
  Original Source: P. Van Hentenryck's book
  Adapted by     : Daniel Diaz - INRIA France
  Date           : September 1992
  '''
"""

def objective(x):
    return 0.0

nletters = 26

bounds = [(1,nletters)]*nletters
# with penalty='penalty' applied, solution is:
#      A   B   C   D   E   F   G   H   I   J   K   L   M   N   O   P   Q
xs = [ 5, 13,  9, 16, 20,  4, 24, 21, 25, 17, 23,  2,  8, 12, 10, 19,  7, \
#      R   S   T   U   V   W   X   Y   Z
      11, 15,  3,  1, 26,  6, 22, 14, 18]
ys = 0.0

# constraints
equations = """
B + A + L + L + E + T - 45 == 0
C + E + L + L + O - 43 == 0
C + O + N + C + E + R + T - 74 == 0
F + L + U + T + E - 30 == 0
F + U + G + U + E - 50 == 0
G + L + E + E - 66 == 0
J + A + Z + Z - 58 == 0
L + Y + R + E - 47 == 0
O + B + O + E - 53 == 0
O + P + E + R + A - 65 == 0
P + O + L + K + A - 59 == 0
Q + U + A + R + T + E + T - 50 == 0
S + A + X + O + P + H + O + N + E - 134 == 0
S + C + A + L + E - 51 == 0
S + O + L + O - 37 == 0
S + O + N + G - 61 == 0
S + O + P + R + A + N + O - 82 == 0
T + H + E + M + E - 72 == 0
V + I + O + L + I + N - 100 == 0
W + A + L + T + Z - 34 == 0
"""
var = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
#NOTE: FOR A MORE DIFFICULT PROBLEM, COMMENT OUT THE FOLLOWING 5 LINES
bounds[0] = (5,5)    # A
bounds[4] = (20,20)  # E
bounds[8] = (25,25)  # I
bounds[14] = (10,10) # O
bounds[20] = (1,1)   # U

from mystic.constraints import unique, near_integers, has_unique

from mystic.symbolic import generate_penalty, generate_conditions
pf = generate_penalty(generate_conditions(equations,var),k=1)
from mystic.constraints import as_constraint
cf = as_constraint(pf)
from mystic.penalty import quadratic_equality

@quadratic_equality(near_integers)
@quadratic_equality(has_unique)
def penalty(x):
    return pf(x)

from numpy import round, hstack, clip
def constraint(x):
    x = round(x).astype(int) # force round and convert type to int
    x = clip(x, 1,nletters)  #XXX: hack to impose bounds
    x = unique(x, list(range(1,nletters+1)))
    return x


if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual
    from mystic.monitors import Monitor, VerboseMonitor
    mon = VerboseMonitor(10)#,10)

    result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=constraint, npop=52, ftol=1e-8, gtol=1000, disp=True, full_output=True, cross=0.1, scale=0.9, itermon=mon)
   # FIXME: solves at 0%... but w/ vowels fixed 80%?
   #result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=constraint, npop=52, ftol=1e-8, gtol=2000, disp=True, full_output=True, cross=0.1, scale=0.9, itermon=mon)
   #result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=constraint, npop=130, ftol=1e-8, gtol=1000, disp=True, full_output=True, cross=0.1, scale=0.9, itermon=mon)
   #result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=constraint, npop=260, ftol=1e-8, gtol=500, disp=True, full_output=True, cross=0.1, scale=0.9, itermon=mon)
   #result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=constraint, npop=520, ftol=1e-8, gtol=100, disp=True, full_output=True, cross=0.1, scale=0.9, itermon=mon)

    print(result[0])
    assert almostEqual(result[0], xs, tol=1e-8)
    assert almostEqual(result[1], ys, tol=1e-4)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Classes for discrete measure data objects.
Includes point_mass, measure, product_measure, and scenario classes.
"""
# Adapted from seesaw2d.py in branches/UQ/math/examples2/ 
# For usage example, see seesaw2d_inf_example.py .

from mystic.math.measures import impose_expectation, impose_expected_variance
from mystic.math.measures import impose_expected_mean_and_variance
from mystic.math.measures import impose_mean, impose_spread, impose_variance
from mystic.math.measures import impose_weight_norm

__all__ = ['point_mass','measure','product_measure','scenario',\
           '_mimic','_uniform_weights','_list_of_measures','compose',\
           'decompose','unflatten','flatten','bounded_mean',\
           'norm_wts_constraintsFactory','mean_y_norm_wts_constraintsFactory',\
           'impose_feasible','impose_valid']

class point_mass(object):
  """a point mass object with weight and position

Args:
    position (tuple(float)): position of the point mass
    weight (float, default=1.0): weight of the point mass

Attributes:
    position (tuple(float)): position of the point mass
    weight (float): weight of the point mass
"""

  def __init__(self, position, weight=1.0):
    self.weight = weight
    self.position = position
    return

  def __repr__(self):
    return "(%s @%s)" % (self.weight, self.position)

  def __rms(self): # square root of sum of squared positions
    from math import sqrt
    return sqrt(sum([i**2 for i in self.position]))

  # interface
  rms = property(__rms, doc='readonly: square root of the sum of squared position')

  pass

class measure(list):  #FIXME: meant to only accept point_masses...
  """a 1-d collection of point masses forming a 'discrete measure'

Args:
    iterable (list): a list of ``mystic.math.discrete.point_mass`` objects

Notes:
    - assumes only contains ``mystic.math.discrete.point_mass`` objects
    - assumes ``measure.n = len(measure.positions) == len(measure.weights)``
    - relies on constraints to impose notions such as ``sum(weights) == 1.0``
"""

  def support_index(self, tol=0):
    """get the indices where there is support (i.e. non-zero weight)

Args:
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the list of indices where there is support
"""
    from .measures import support_index
    return support_index(self.weights, tol)

  def support(self, tol=0):
    """get the positions with non-zero weight (i.e. support)

Args:
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the list of positions with support
"""
    from .measures import support
    return support(self.positions, self.weights, tol)

  def __weights(self):
    return [i.weight for i in self]

  def __positions(self):
    return [i.position for i in self]

  def __n(self):
    return len(self)

  def __mass(self):
    return sum(self.weights)
    #from mystic.math.measures import norm
    #return norm(self.weights)  # normalized by self.npts

  def __mean(self):
    from mystic.math.measures import mean
    return mean(self.positions, self.weights) 

  def __range(self):
    from mystic.math.measures import spread
    return spread(self.positions)

  def __variance(self):
    from mystic.math.measures import variance
    return variance(self.positions, self.weights)

  def __set_weights(self, weights):
    for i in range(len(weights)):
      self[i].weight = weights[i]
    return

  def __set_positions(self, positions):
    for i in range(len(positions)):
      self[i].position = positions[i]
    return

  def normalize(self):
    """normalize the weights to 1.0

Args:
    None

Returns:
    None
"""
    self.positions, self.weights = impose_weight_norm(self.positions, self.weights)
    return

  def __set_mean(self, m):
    self.positions = impose_mean(m, self.positions, self.weights)
    return

  def __set_range(self, r):
    self.positions = impose_spread(r, self.positions, self.weights)
    return

  def __set_variance(self, v):
    self.positions = impose_variance(v, self.positions, self.weights)
    return

  def maximum(self, f):
    """calculate the maximum for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the maximum value of ``f`` over all measure positions
"""
    from .measures import maximum
    positions = [(i,) for i in self.positions]
    return maximum(f, positions)

  def ess_maximum(self, f, tol=0.):
    """calculate the maximum for the support of a given function

Args:
    f (func): a function that takes a list and returns a number
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the maximum value of ``f`` over all measure positions with support
"""
    from .measures import ess_maximum
    positions = [(i,) for i in self.positions]
    return ess_maximum(f, positions, self.weights, tol)

  def minimum(self, f):
    """calculate the minimum for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the minimum value of ``f`` over all measure positions
"""
    from .measures import minimum
    positions = [(i,) for i in self.positions]
    return minimum(f, positions)

  def ess_minimum(self, f, tol=0.):
    """calculate the minimum for the support of a given function

Args:
    f (func): a function that takes a list and returns a number
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the minimum value of ``f`` over all measure positions with support
"""
    from .measures import ess_minimum
    positions = [(i,) for i in self.positions]
    return ess_minimum(f, positions, self.weights, tol)

  def ptp(self, f):
    """calculate the spread for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the spread of the values of ``f`` over all measure positions
"""
    from .measures import ptp
    positions = [(i,) for i in self.positions]
    return ptp(f, positions)

  def ess_ptp(self, f, tol=0.):
    """calculate the spread for the support of a given function

Args:
    f (func): a function that takes a list and returns a number
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the spread of the values of ``f`` over all measure positions with support
"""
    from .measures import ess_ptp
    positions = [(i,) for i in self.positions]
    return ess_ptp(f, positions, self.weights, tol)

  def expect(self, f):
    """calculate the expectation for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the expectation of ``f`` over all measure positions
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?
    from mystic.math.measures import expectation
    positions = [(i,) for i in self.positions]
    return expectation(f, positions, self.weights)

  def expect_var(self, f):
    """calculate the expected variance for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the expected variance of ``f`` over all measure positions
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?
    from mystic.math.measures import expected_variance
    positions = [(i,) for i in self.positions]
    return expected_variance(f, positions, self.weights)

  #NOTE: backward incompatible 08/26/18: (expected=(m,D),...) --> (m,...,tol=D)
  def set_expect(self, expected, f, bounds=None, constraints=None, **kwds):
    """impose an expectation on the measure by adjusting the positions

Args:
    expected (float): target expected mean
    f (func): a function that takes a list and returns a number
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``c' = constraints(c)``,
        where ``c`` is a product measure, and ``c'`` is a product measure
        where the encoded constaints are satisfied.
    tol (float, default=None): maximum allowable deviation from ``expected``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    None

Notes:
    Expectation ``E`` is calculated by minimizing ``mean(f(x)) - expected``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target mean ``expected``. If ``tol`` is not
    provided, then a relative deviation of 1% of ``expected`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from.

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter.
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?
    #XXX: maybe also natural c' = constraints(c) where c is a measure ?
    m = expected
    if constraints:  # then adjust interface for 'impose_expectation'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    positions = impose_expectation(m, f, [self.npts], bounds, self.weights, \
                                   constraints=cnstr, **kwds)
    from numpy import array
    self.positions = list(array(positions)[:,0])
   #from numpy import squeeze
   #self.positions = list(squeeze(positions))
    return

  def set_expect_var(self, expected, f, bounds=None, constraints=None, **kwds):
    """impose an expected variance on the measure by adjusting the positions

Args:
    expected (float): target expected variance
    f (func): a function that takes a list and returns a number
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``c' = constraints(c)``,
        where ``c`` is a product measure, and ``c'`` is a product measure
        where the encoded constaints are satisfied.
    tol (float, default=None): maximum allowable deviation from ``expected``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    None

Notes:
    Expected var ``E`` is calculated by minimizing ``var(f(x)) - expected``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target variance ``expected``. If ``tol`` is not
    provided, then a relative deviation of 1% of ``expected`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from.

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter.
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?
    #XXX: maybe also natural c' = constraints(c) where c is a measure ?
    m = expected
    if constraints:  # then adjust interface for 'impose_expected_variance'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    positions = impose_expected_variance(m, f, [self.npts], bounds, \
                                self.weights, constraints=cnstr, **kwds)
    from numpy import array
    self.positions = list(array(positions)[:,0])
   #from numpy import squeeze
   #self.positions = list(squeeze(positions))
    return

  def set_expect_mean_and_var(self, expected, f, bounds=None, constraints=None, **kwds):
    """impose expected mean and var on the measure by adjusting the positions

Args:
    expected (tuple(float)): ``(expected mean, expected var)``
    f (func): a function that takes a list and returns a number
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``c' = constraints(c)``,
        where ``c`` is a product measure, and ``c'`` is a product measure
        where the encoded constaints are satisfied.
    tol (float, default=None): maximum allowable deviation from ``expected``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    None

Notes:
    Expected mean ``E`` and expected variance ``R`` are calculated by
    minimizing the sum of the absolute values of ``mean(f(x)) - m`` and
    ``variance(f(x)) - v`` over the given *bounds*, and will terminate when
    ``E`` and ``R`` are found within tolerance ``tol`` of the target mean ``m``
    and variance ``v``, respectively. If ``tol`` is not provided, then a
    relative deviation of 1% of ``max(m,v)`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter
""" #XXX: maybe more natural if f takes a positional value x, not a list x ?
    #XXX: maybe also natural c' = constraints(c) where c is a measure ?
    m,v = expected
    if constraints:  # then adjust interface for 'impose_expected_mean_and_var'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    positions = impose_expected_mean_and_variance((m,v), f, [self.npts], \
                       bounds, self.weights, constraints=cnstr, **kwds)
    from numpy import array
    self.positions = list(array(positions)[:,0])
   #from numpy import squeeze
   #self.positions = list(squeeze(positions))
    return

  # interface
  weights = property(__weights, __set_weights, doc='a list of weights for all point masses in the measure')
  positions = property(__positions, __set_positions, doc='a list of positions for all point masses in the measure')
  ###XXX: why not use 'points' also/instead?
  npts = property(__n, doc='readonly: the number of point masses in the measure')
  mass = property(__mass, doc='readonly: the sum of the weights')
  range = property(__range, __set_range, doc='``|max - min|`` for the positions')
  center_mass = property(__mean, __set_mean, doc='sum of ``weights * positions``')
  var = property(__variance, __set_variance, doc='``mean(|positions - mean(positions)|**2)``')

  # backward compatibility
  coords = positions
  get_expect = expect
  mean = center_mass
  pass

class product_measure(list):  #FIXME: meant to only accept sets...
  """a N-d measure-theoretic product of discrete measures

Args:
    iterable (list): a list of ``mystic.math.discrete.measure`` objects

Notes:
    - all measures are treated as if they are orthogonal
    - assumes only contains ``mystic.math.discrete.measure`` objects
    - assumes ``len(product_measure.positions) == len(product_measure.weights)``
    - relies on constraints to impose notions such as ``sum(weights) == 1.0``
    - relies on constraints to impose expectation (within acceptable deviation)
    - positions are ``(xi,yi,zi)`` with weights ``(wxi,wyi,wzi)``, where weight
      ``wxi`` at ``xi`` should be the same for each ``(yj,zk)``.  Similarly
      for each ``wyi`` and ``wzi``.
"""
  def __val(self):
    raise NotImplementedError("'value' is undefined in a measure")

  def __pts(self):
    return [i.npts for i in self]

  def __wts(self):
    return [i.weights for i in self]

  def __pos(self):
    return [i.positions for i in self]

  def __mean(self):
    return [i.center_mass for i in self]

  def __set_mean(self, center_masses):
    [i._measure__set_mean(center_masses[m]) for (m,i) in enumerate(self)]
   #for i in range(len(center_masses)):
   #  self[i].center_mass = center_masses[i]
    return

  def __n(self):
    from numpy import product
    return product(self.pts)

  def support_index(self, tol=0):
    """get the indices where there is support (i.e. non-zero weight)

Args:
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the list of indices where there is support
"""
    from .measures import support_index
    return support_index(self.weights, tol)

  def support(self, tol=0): #XXX: better if generated positions only when needed
    """get the positions with non-zero weight (i.e. support)

Args:
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the list of positions with support
"""
    from .measures import support
    return support(self.positions, self.weights, tol)

  def __weights(self):
    from mystic.math.measures import _pack
    from numpy import product
    weights = _pack(self.wts)
    _weights = []
    for wts in weights:
      _weights.append(product(wts))
    return _weights

  def __positions(self):
    from mystic.math.measures import _pack
    return _pack(self.pos)

  def __set_positions(self, positions):
    from mystic.math.measures import _unpack
    positions = _unpack(positions, self.pts)
    for i in range(len(positions)):
      self[i].positions = positions[i]
    return

 #def __get_center(self):
 #  return self.__center

 #def __get_delta(self):
 #  return self.__delta

  def __mass(self):
    return [self[i].mass for i in range(len(self))]

  def maximum(self, f): #XXX: return max of all or return all max?
    """calculate the maximum for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the maximum value of ``f`` over all measure positions
"""
    return max([i.maximum(f) for i in self])

  def minimum(self, f): #XXX: return min of all or return all min?
    """calculate the minimum for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the minimum value of ``f`` over all measure positions
"""
    return min([i.minimum(f) for i in self])

  def ptp(self, f): #XXX: return max of all or return all ptp?
    """calculate the spread for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the spread for values of ``f`` over all measure positions
"""
    return max([i.ptp(f) for i in self])

  def ess_maximum(self, f, tol=0.): #XXX: return max of all or return all max?
    """calculate the maximum for the support of a given function

Args:
    f (func): a function that takes a list and returns a number
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the maximum value of ``f`` over all measure positions with support
"""
    return max([i.ess_maximum(f, tol) for i in self])

  def ess_minimum(self, f, tol=0.): #XXX: return min of all or return all min?
    """calculate the minimum for the support of a given function

Args:
    f (func): a function that takes a list and returns a number
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the minimum value of ``f`` over all measure positions with support
"""
    return min([i.ess_minimum(f, tol) for i in self])

  def ess_ptp(self, f, tol=0.): #XXX: return max of all or return all ptp?
    """calculate the spread for the support of a given function

Args:
    f (func): a function that takes a list and returns a number
    tol (float, default=0.0): tolerance, where any ``weight <= tol`` is zero

Returns:
    the spread of values of ``f`` over all measure positions with support
"""
    return max([i.ess_ptp(f, tol) for i in self])

  def expect(self, f):
    """calculate the expectation for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the expectation of ``f`` over all measure positions
"""
    from mystic.math.measures import expectation
    return expectation(f, self.positions, self.weights)

  def expect_var(self, f):
    """calculate the expected variance for a given function

Args:
    f (func): a function that takes a list and returns a number

Returns:
    the expected variance of ``f`` over all measure positions
"""
    from mystic.math.measures import expected_variance
    return expected_variance(f, self.positions, self.weights)

  #NOTE: backward incompatible 08/26/18: (expected=(m,D),...) --> (m,...,tol=D)
  def set_expect(self, expected, f, bounds=None, constraints=None, **kwds):
    """impose an expectation on the measure by adjusting the positions

Args:
    expected (float): target expected mean
    f (func): a function that takes a list and returns a number
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``c' = constraints(c)``,
        where ``c`` is a product measure, and ``c'`` is a product measure
        where the encoded constaints are satisfied.
    tol (float, default=None): maximum allowable deviation from ``expected``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    None

Notes:
    Expectation ``E`` is calculated by minimizing ``mean(f(x)) - expected``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target mean ``expected``. If ``tol`` is not
    provided, then a relative deviation of 1% of ``expected`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from.

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter.
"""
    #self.__center = m
    #self.__delta = D
    m = expected
    if constraints:  # then adjust interface for 'impose_expectation'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    self.positions = impose_expectation(m, f, self.pts, bounds, self.weights, \
                                        constraints=cnstr, **kwds)
    return

  def set_expect_var(self, expected, f, bounds=None, constraints=None, **kwds):
    """impose an expected variance on the measure by adjusting the positions

Args:
    expected (float): target expected variance
    f (func): a function that takes a list and returns a number
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``c' = constraints(c)``,
        where ``c`` is a product measure, and ``c'`` is a product measure
        where the encoded constaints are satisfied.
    tol (float, default=None): maximum allowable deviation from ``expected``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    None

Notes:
    Expected var ``E`` is calculated by minimizing ``var(f(x)) - expected``,
    over the given *bounds*, and will terminate when ``E`` is found within
    deviation ``tol`` of the target variance ``expected``. If ``tol`` is not
    provided, then a relative deviation of 1% of ``expected`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from.

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter.
"""
    #self.__center = m
    #self.__delta = D
    m = expected
    if constraints:  # then adjust interface for 'impose_expected_variance'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    self.positions = impose_expected_variance(m, f, self.pts, bounds, \
                             self.weights, constraints=cnstr, **kwds)
    return


  def set_expect_mean_and_var(self, expected, f, bounds=None, constraints=None, **kwds):
    """impose expected mean and var on the measure by adjusting the positions

Args:
    expected (tuple(float)): ``(expected mean, expected var)``
    f (func): a function that takes a list and returns a number
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``c' = constraints(c)``,
        where ``c`` is a product measure, and ``c'`` is a product measure
        where the encoded constaints are satisfied.
    tol (float, default=None): maximum allowable deviation from ``expected``
    npop (int, default=200): size of the trial solution population
    maxiter (int, default=1000): the maximum number of iterations to perform
    maxfun (int, default=1e+6): the maximum number of function evaluations

Returns:
    None

Notes:
    Expected mean ``E`` and expected variance ``R`` are calculated by
    minimizing the sum of the absolute values of ``mean(f(x)) - m`` and
    ``variance(f(x)) - v`` over the given *bounds*, and will terminate when
    ``E`` and ``R`` are found within tolerance ``tol`` of the target mean ``m``
    and variance ``v``, respectively. If ``tol`` is not provided, then a
    relative deviation of 1% of ``max(m,v)`` will be used.

    This function does not preserve the mean, variance, or range, as there
    is no initial list of samples to draw the mean, variance, and etc from

    *bounds* is tuple with ``length(bounds) == 2``, composed of all the lower
    bounds, then all the upper bounds, for each parameter
"""
    #self.__center = m
    #self.__delta = D
    m,v = expected
    if constraints:  # then adjust interface for 'impose_expected_mean_and_var'
      def cnstr(x, w):
        c = compose(x,w)
        c = constraints(c)
        return decompose(c)[0]
    else: cnstr = constraints  # 'should' be None
    self.positions = impose_expected_mean_and_variance((m,v), f, self.pts, \
                          bounds, self.weights, constraints=cnstr, **kwds)
    return


  def pof(self, f):
    """calculate probability of failure for a given function

Args:
    f (func): a function returning True for 'success' and False for 'failure'

Returns:
    the probabilty of failure, a float in ``[0.0,1.0]``

Notes:
    - the function ``f`` should take a list of ``positions`` (for example,
      ``scenario.positions`` or ``product_measure.positions``) and return a
      single value (e.g. 0.0 or False)
"""
    u = 0.0
    set = zip(self.positions, self.weights)
    for x in set:
      if f(x[0]) <= 0.0: #XXX: should f return a bool or a float?
        u += x[1]
    return u
  # for i in range(self.npts):
  #   #if f(self.positions[i]) > 0.0:  #NOTE: f(x) > 0.0 yields prob of success
  #   if f(self.positions[i]) <= 0.0:  #NOTE: f(x) <= 0.0 yields prob of failure
  #     u += self.weights[i]
  # return u  #XXX: does this need to be normalized?

  def sampled_minimum(self, f, npts=10000):
    """use sampling to calculate ess_minimum for a given function

Args:
    f (func): a function that takes a list and returns a number
    npts (int, default=10000): the number of point masses sampled from the
        underlying discrete measures

Returns:
    the sampled ess_minimum, a float

Notes:
    - the function ``f`` should take a list of ``positions`` (for example,
      ``scenario.positions`` or ``product_measure.positions``) and return a
      single value (e.g. 0.0)
"""
    from mystic.math.samples import _minimum_given_samples
    pts = self.sampled_support(npts)
    return _minimum_given_samples(f, pts)

  def sampled_ptp(self, f, npts=10000):
    """use sampling to calculate ess_|maximum - minimum| for a given function

Args:
    f (func): a function that takes a list and returns a number
    npts (int, default=10000): the number of point masses sampled from the
        underlying discrete measures

Returns:
    the sampled |ess_maximum - ess_minimum|, a float

Notes:
    - the function ``f`` should take a list of ``positions`` (for example,
      ``scenario.positions`` or ``product_measure.positions``) and return a
      single value (e.g. 0.0)
"""
    from mystic.math.samples import _ptp_given_samples
    pts = self.sampled_support(npts)
    return _ptp_given_samples(f, pts)

  def sampled_expect(self, f, npts=10000):
    """use sampling to calculate expected value for a given function

Args:
    f (func): a function that takes a list and returns a number
    npts (int, default=10000): the number of point masses sampled from the
        underlying discrete measures

Returns:
    the expected value, a float

Notes:
    - the function ``f`` should take a list of ``positions`` (for example,
      ``scenario.positions`` or ``product_measure.positions``) and return a
      single value (e.g. 0.0)
"""
    from mystic.math.samples import _expectation_given_samples
    pts = self.sampled_support(npts)
    return _expectation_given_samples(f, pts)

  def sampled_maximum(self, f, npts=10000):
    """use sampling to calculate ess_maximum for a given function

Args:
    f (func): a function that takes a list and returns a number
    npts (int, default=10000): the number of point masses sampled from the
        underlying discrete measures

Returns:
    the ess_maximum, a float

Notes:
    - the function ``f`` should take a list of ``positions`` (for example,
      ``scenario.positions`` or ``product_measure.positions``) and return a
      single value (e.g. 0.0)
"""
    from mystic.math.samples import _maximum_given_samples
    pts = self.sampled_support(npts)
    return _maximum_given_samples(f, pts)

  def sampled_pof(self, f, npts=10000):
    """use sampling to calculate probability of failure for a given function

Args:
    f (func): a function returning True for 'success' and False for 'failure'
    npts (int, default=10000): the number of point masses sampled from the
        underlying discrete measures

Returns:
    the probabilty of failure, a float in ``[0.0,1.0]``

Notes:
    - the function ``f`` should take a list of ``positions`` (for example,
      ``scenario.positions`` or ``product_measure.positions``) and return a
      single value (e.g. 0.0 or False)
"""
    from mystic.math.samples import _pof_given_samples
    pts = self.sampled_support(npts)
    return _pof_given_samples(f, pts)

  def sampled_support(self, npts=10000): ##XXX: was 'def support'
    """randomly select support points from the underlying discrete measures

Args:
    npts (int, default=10000): the number of sampled points

Returns:
    a list of ``len(product measure)`` lists, each of length ``len(npts)``
"""
    from mystic.math.measures import weighted_select as _select
    pts = []
    for i in range(npts):
      # for a single trial, select positions from all sets
      pts.append( [_select(set.positions, set.weights) for set in self] )

    # convert pts to len(prod_meas) lists, each of len(npts)
    from numpy import transpose
    return transpose(pts)  #XXX: assumes 'positions' is a list of floats

  def update(self, params):
    """update the product measure from a list of parameters

Args:
    params (list(float)): parameters corresponding to N 1D discrete measures

Returns:
    self (measure): the product measure itself

Notes:
    The dimensions of the product measure will not change upon update, and
    it is assumed *params* either corresponds to the correct number of
    weights and positions for the existing ``product_measure``, or *params*
    has additional values (typically output values) which will be ignored.
    It is assumed that ``len(params) >= 2 * sum(product_measure.pts)``.

    If ``product_measure.pts = (M, N, ...)``, then it is assumed that
    ``params = [wx1, ..., wxM, x1, ..., xM, wy1, ..., wyN, y1, ..., yN, ...]``.
    Thus, *params* should have ``M`` weights and ``M`` corresponding positions,
    followed by ``N`` weights and ``N`` corresponding positions, with this
    pattern followed for each new dimension of the desired product measure.
"""
    pts = self.pts
    _len = 2 * sum(pts)

    if len(params)  >  _len:  # if Y-values are appended to params
      params, values  =  params[:_len], params[_len:]

    pm = unflatten(params, pts)
    zo = pm.count([])
    self[:] = pm[:len(self) - zo] + self[len(pm) - zo:]
    return self

  def load(self, params, pts):
    """load the product measure from a list of parameters

Args:
    params (list(float)): parameters corresponding to N 1D discrete measures
    pts (tuple(int)): number of point masses in each of the discrete measures

Returns:
    self (measure): the product measure itself

Notes:
    To append ``len(pts)`` new discrete measures to the product measure,
    it is assumed *params* either corresponds to the correct number of
    weights and positions specified by *pts*, or *params* has additional
    values (typically output values) which will be ignored. It is assumed
    that ``len(params) >= 2 * sum(product_measure.pts)``.

    Given the value of ``pts = (M, N, ...)``, it is assumed that
    ``params = [wx1, ..., wxM, x1, ..., xM, wy1, ..., wyN, y1, ..., yN, ...]``.
    Thus, *params* should have ``M`` weights and ``M`` corresponding positions,
    followed by ``N`` weights and ``N`` corresponding positions, with this
    pattern followed for each new dimension of the desired product measure.
"""
    _len = 2 * sum(pts)
    if len(params)  >  _len:  # if Y-values are appended to params
      params, values  =  params[:_len], params[_len:]

    self.extend( unflatten(params, pts) )
    return self

  def flatten(self):
    """convert a product measure to a single list of parameters

Args:
    None

Returns:
    a list of parameters

Notes:
    Given ``product_measure.pts = (M, N, ...)``, then the returned list is
    ``params = [wx1, ..., wxM, x1, ..., xM, wy1, ..., wyN, y1, ..., yN, ...]``.
    Thus, *params* will have ``M`` weights and ``M`` corresponding positions,
    followed by ``N`` weights and ``N`` corresponding positions, with this
    pattern followed for each new dimension of the desired product measure.
"""
    params = flatten(self)
    return params

  #XXX: name stinks... better as "non_redundant"? ...is really a helper
  def differs_by_one(self, ith, all=True, index=True):
    """get the coordinates where the associated binary string differs
by exactly one index

Args:
    ith (int): the target index
    all (bool, default=True): if False, only return results for indices < ``i``
    index (bool, default=True): if True, return the indices of the results
        instead of the results themselves

Returns:
    the coordinates where the associated binary string differs by one, or
    if *index* is True, return the corresponding indices
"""
    from mystic.math.compressed import index2binary, differs_by_one
    b = index2binary(list(range(self.npts)), self.npts)
    return differs_by_one(ith, b, all, index) 

  def select(self, *index, **kwds):
    """generate product measure positions for the selected position indices

Args:
    index (tuple(int)): tuple of position indicies

Returns:
    a list of product measure positions for the selected indices

Examples:
    >>> r
    [[9, 8], [1, 3], [4, 2]]
    >>> r.select(*range(r.npts))
    [(9, 1, 4), (8, 1, 4), (9, 3, 4), (8, 3, 4), (9, 1, 2), (8, 1, 2), (9, 3, 2), (8, 3, 2)]
    >>>
    >>> _pack(r)
    [(9, 1, 4), (8, 1, 4), (9, 3, 4), (8, 3, 4), (9, 1, 2), (8, 1, 2), (9, 3, 2), (8, 3, 2)]

Notes:
    This only works for product measures of dimension ``2^K``
"""
    from mystic.math.compressed import index2binary, binary2coords
    v = index2binary(list(index), self.npts)
    return binary2coords(v, self.pos, **kwds)
    #XXX: '_pack' requires resorting ([::-1]) so that indexing is wrong.
    #     Better if modify mystic's pack to match sorting of binary strings ?

 #__center = None
 #__delta = None

  # interface
  npts = property(__n, doc='readonly: the total number of point masses in the product measure')
  weights = property(__weights, doc='a list of weights for all point masses in the product measure')
  positions = property(__positions, __set_positions, doc='a list of positions for all point masses in the product measure')
  center_mass = property(__mean, __set_mean, doc='sum of ``weights * positions``')
 #center = property(__get_center ) #FIXME: remove c.center and c.delta... or
 #delta = property(__get_delta )   #       replace with c._params (e.g. (m,D))
 #expect = property(__expect, __set_expect )
  mass = property(__mass, doc='readonly: a list of weight norms')
  pts = property(__pts, doc='readonly: the number of point masses for each discrete mesure')
  wts = property(__wts, doc='readonly: a list of weights for each discrete mesure')
  pos = property(__pos, doc='readonly: a list of positions for each discrete mesure')

  # backward compatibility
  coords = positions
  get_expect = expect
  pass


class scenario(product_measure):  #FIXME: meant to only accept sets...
  """a N-d product measure with associated data values

A scenario is a measure-theoretic product of discrete measures that also
includes a list of associated values, with the values corresponding to
measured or synthetic data for each measure position.  Each point mass
in the product measure is paired with a value, and thus, essentially, a
scenario is equivalent to a ``mystic.math.legacydata.dataset`` stored in 
a ``product_measure`` representation.

Args:
    pm (mystic.math.discrete.product_measure, default=None): a product measure
    values (list(float), default=None): values associated with each position

Notes:
    - all measures are treated as if they are orthogonal
    - relies on constraints to impose notions such as ``sum(weights) == 1.0``
    - relies on constraints to impose expectation (within acceptable deviation)
    - positions are ``(xi,yi,zi)`` with weights ``(wxi,wyi,wzi)``, where weight
      ``wxi`` at ``xi`` should be the same for each ``(yj,zk)``.  Similarly
      for each ``wyi`` and ``wzi``.
"""
  def __init__(self, pm=None, values=None):
    super(product_measure,self).__init__()
    if pm: 
      pm = product_measure(pm)
      self.load(pm.flatten(), pm.pts)
    if not values: values = []
    self.__Y = values # storage for values of s.positions
    return

  def __values(self):
    return self.__Y

  def __set_values(self, values):
    self.__Y = values[:]
    return

  def mean_value(self):  # get mean of y's
    """calculate the mean of the associated values for a scenario

Args:
    None

Returns:
    the weighted mean of the scenario values
"""
    from mystic.math.measures import mean
    return mean(self.values, self.weights)

  def set_mean_value(self, m):  # set mean of y's
    """set the mean for the associated values of a scenario

Args:
    m (float): the target weighted mean of the scenario values

Returns:
    None
"""
    from mystic.math.measures import impose_mean
    self.values = impose_mean(m, self.values, self.weights)
    return

  def valid_wrt_model(self, model, blamelist=False, pairs=True, \
                                   all=False, raw=False, **kwds):
    """check for scenario validity with respect to the model

Args:
    model (func): the model function, ``y' = F(x')``.
    blamelist (bool, default=False): if True, indicate the infeasible points.
    pairs (bool, default=True): if True, indicate indices of infeasible points.
    all (bool, default=False): if True, get results for each individual point.
    raw (bool, default=False): if False, get boolean results (i.e. non-float).
    ytol (float, default=0.0): maximum acceptable difference ``|y - F(x')|``.
    xtol (float, default=0.0): maximum acceptable difference ``|x - x'|``.
    cutoff (float, default=ytol): zero out distances less than cutoff.
    hausdorff (bool, default=False): hausdorff ``norm``, where if given,
        then ``ytol = |y - F(x')| + |x - x'|/norm``.

Notes:
    *xtol* defines the n-dimensional base of a pilar of height *ytol*,
    centered at each point. The region inside the pilar defines the space
    where a "valid" model must intersect. If *xtol* is not specified, then
    the base of the pilar will be a dirac at ``x' = x``. This function
    performs an optimization for each ``x`` to find an appropriate ``x'``.

    *ytol* is a single value, while *xtol* is a single value or an iterable.
    *cutoff* takes a float or a boolean, where ``cutoff=True`` will set the
    value of *cutoff* to the default. Typically, the value of *cutoff* is
    *ytol*, 0.0, or None. *hausdorff* can be False (e.g. ``norm = 1.0``),
    True (e.g. ``norm = spread(x)``), or a list of points of ``len(x)``.

    While *cutoff* and *ytol* are very tightly related, they play a distinct
    role; *ytol* is used to set the optimization termination for an acceptable
    ``|y - F(x')|``, while *cutoff* is applied post-optimization.

    If we are using the *hausdorff* norm, then *ytol* will set the optimization
    termination for an acceptable ``|y - F(x')| + |x - x'|/norm``, where the
    ``x`` values are normalized by ``norm = hausdorff``.
"""
    from mystic.math.legacydata import dataset 
    data = dataset() 
    data.load(self.positions, self.values)
   #data.lipschitz = L
    for i in range(len(data)):
      data[i].id = i
    return data.valid(model, blamelist=blamelist, pairs=pairs, \
                                       all=all, raw=raw, **kwds)

  def short_wrt_self(self, L, blamelist=False, pairs=True, \
                              all=False, raw=False, **kwds):
    """check for shortness with respect to the scenario itself

Args:
    L (float): the lipschitz constant.
    blamelist (bool, default=False): if True, indicate the infeasible points.
    pairs (bool, default=True): if True, indicate indices of infeasible points.
    all (bool, default=False): if True, get results for each individual point.
    raw (bool, default=False): if False, get boolean results (i.e. non-float).
    tol (float, default=0.0): maximum acceptable deviation from shortness.
    cutoff (float, default=tol): zero out distances less than cutoff.

Notes:
    Each point x,y can be thought to have an associated double-cone with slope
    equal to the lipschitz constant. Shortness with respect to another point is
    defined by the first point not being inside the cone of the second. We can
    allow for some error in shortness, a short tolerance *tol*, for which the
    point x,y is some acceptable y-distance inside the cone. While very tightly
    related, *cutoff* and *tol* play distinct roles; *tol* is subtracted from
    calculation of the lipschitz_distance, while *cutoff* zeros out the value
    of any element less than the *cutoff*.
"""
    from mystic.math.legacydata import dataset 
    data = dataset() 
    data.load(self.positions, self.values)
    data.lipschitz = L
    for i in range(len(data)):
      data[i].id = i
    return data.short(blamelist=blamelist, pairs=pairs, \
                                           all=all, raw=raw, **kwds)

  def short_wrt_data(self, data, L=None, blamelist=False, pairs=True, \
                                         all=False, raw=False, **kwds):
    """check for shortness with respect to the given data

Args:
    data (list): a list of data points or dataset to compare against.
    L (float, default=None): the lipschitz constant, if different than in data.
    blamelist (bool, default=False): if True, indicate the infeasible points.
    pairs (bool, default=True): if True, indicate indices of infeasible points.
    all (bool, default=False): if True, get results for each individual point.
    raw (bool, default=False): if False, get boolean results (i.e. non-float).
    tol (float, default=0.0): maximum acceptable deviation from shortness.
    cutoff (float, default=tol): zero out distances less than cutoff.

Notes:
    Each point x,y can be thought to have an associated double-cone with slope
    equal to the lipschitz constant. Shortness with respect to another point is
    defined by the first point not being inside the cone of the second. We can
    allow for some error in shortness, a short tolerance *tol*, for which the
    point x,y is some acceptable y-distance inside the cone. While very tightly
    related, *cutoff* and *tol* play distinct roles; *tol* is subtracted from
    calculation of the lipschitz_distance, while *cutoff* zeros out the value
    of any element less than the *cutoff*.
"""
    from mystic.math.legacydata import dataset 
    _self = dataset() 
    _self.load(self.positions, self.values)
    _self.lipschitz = data.lipschitz
    for i in range(len(_self)):
      _self[i].id = i
    return _self.short(data, L=L, blamelist=blamelist, pairs=pairs, \
                                               all=all, raw=raw, **kwds)

  def set_feasible(self, data, cutoff=0.0, bounds=None, constraints=None, \
                                                  with_self=True, **kwds):
    """impose shortness with respect to the given data points

This function attempts to minimize the infeasibility between observed *data*
and the scenario of synthetic data by perforing an optimization on ``w,x,y``
over the given *bounds*.

Args:
    data (mystic.math.discrete.scenario): a dataset of observed points
    cutoff (float, default=0.0): maximum acceptable deviation from shortness
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``x' = constraints(x)``,
        where ``x`` is a scenario that has been converted into a list of
        parameters (e.g. with ``scenario.flatten``), and ``x'`` is the list
        of parameters after the encoded constaints have been satisfied.
    with_self (bool, default=True): if True, shortness is also self-consistent
    tol (float, default=0.0): maximum acceptable optimizer termination
        for ``sum(infeasibility)``.

Returns:
    None

Notes:
    - both ``scenario.positions`` and ``scenario.values`` may be adjusted.
    - if *with_self* is True, shortness will be measured not only from the
      scenario to the given *data*, but also between scenario datapoints.
"""
    # imposes: is_short(x, x'), is_short(x, z )
    # use additional 'constraints' kwds to impose: y >= m, norm(wi) = 1.0
    pm = impose_feasible(cutoff, data, guess=self.pts, bounds=bounds, \
                         constraints=constraints, with_self=with_self, **kwds)
    self.update( pm.flatten(all=True) )
    return

  def set_valid(self, model, cutoff=0.0, bounds=None, constraints=None, **kwds):
    """impose model validity on a scenario by adjusting positions and values

This function attempts to minimize the graph distance between reality (data),
``y = G(x)``, and an approximating function, ``y' = F(x')``, by perforing an
optimization on ``w,x,y`` over the given *bounds*.

Args:
    model (func): a model ``y' = F(x')`` that approximates reality ``y = G(x)``
    cutoff (float, default=0.0): acceptable model invalidity ``|y - F(x')|``
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``x' = constraints(x)``,
        where ``x`` is a scenario that has been converted into a list of
        parameters (e.g. with ``scenario.flatten``), and ``x'`` is the list
        of parameters after the encoded constaints have been satisfied.
    hausdorff (bool, default=False): hausdorff ``norm``, where if given,
        then ``ytol = |y - F(x')| + |x - x'|/norm``
    xtol (float, default=0.0): maximum acceptable pointwise graphical distance
        between model and reality.
    tol (float, default=0.0): maximum acceptable optimizer termination
        for ``sum(graphical distances)``.

Returns:
    None

Notes:
    *xtol* defines the n-dimensional base of a pilar of height *cutoff*,
    centered at each point. The region inside the pilar defines the space
    where a "valid" model must intersect. If *xtol* is not specified, then
    the base of the pilar will be a dirac at ``x' = x``. This function
    performs an optimization to find a set of points where the model is valid.
    Here, *tol* is used to set the optimization termination for minimizing the
    ``sum(graphical_distances)``, while *cutoff* is used in defining the
    graphical distance between ``x,y`` and ``x',F(x')``.
"""
    # imposes is_feasible(R, Cv), where R = graphical_distance(model, pts)
    # use additional 'constraints' kwds to impose: y >= m, norm(wi) = 1.0
    pm = impose_valid(cutoff, model, guess=self, \
                      bounds=bounds, constraints=constraints, **kwds)
    self.update( pm.flatten(all=True) )
    return
 
  def pof_value(self, f):
    """calculate probability of failure for a given function

Args:
    f (func): a function returning True for 'success' and False for 'failure'

Returns:
    the probabilty of failure, a float in ``[0.0,1.0]``

Notes:
    - the function ``f`` should take a list of ``values`` (for example,
      ``scenario.values``) and return a single value (e.g. 0.0 or False)
"""
    u = 0.0
    set = zip(self.values, self.weights)
    for x in set:
      if f(x[0]) <= 0.0:
        u += x[1]
    return u

  def update(self, params): #XXX: overwritten.  create standalone instead ?
    """update the scenario from a list of parameters

Args:
    params (list(float)): parameters corresponding to N 1D discrete measures

Returns:
    self (scenario): the scenario itself

Notes:
    The dimensions of the scenario will not change upon update, and
    it is assumed *params* either corresponds to the correct number of
    weights and positions for the existing ``scenario``, or *params*
    has additional values which will be saved as the ``scenario.values``.
    It is assumed that ``len(params) >= 2 * sum(scenario.pts)``.

    If ``scenario.pts = (M, N, ...)``, then it is assumed that
    ``params = [wx1, ..., wxM, x1, ..., xM, wy1, ..., wyN, y1, ..., yN, ...]``.
    Thus, *params* should have ``M`` weights and ``M`` corresponding positions,
    followed by ``N`` weights and ``N`` corresponding positions, with this
    pattern followed for each new dimension of the desired scenario.
"""
    pts = self.pts
    _len = 2 * sum(pts)

    if len(params)  >  _len:  # if Y-values are appended to params
      params, values  =  params[:_len], params[_len:]
      self.values = values[:len(self.values)] + self.values[len(values):] 

    pm = unflatten(params, pts)
    zo = pm.count([])
    self[:] = pm[:len(self) - zo] + self[len(pm) - zo:]
    return self

  def load(self, params, pts): #XXX: overwritten.  create standalone instead ?
    """load the scenario from a list of parameters

Args:
    params (list(float)): parameters corresponding to N 1D discrete measures
    pts (tuple(int)): number of point masses in each of the discrete measures

Returns:
    self (scenario): the scenario itself

Notes:
    To append ``len(pts)`` new discrete measures to the scenario,
    it is assumed *params* either corresponds to the correct number of
    weights and positions specified by *pts*, or *params* has additional
    values which will be saved as the ``scenario.values``. It is assumed
    that ``len(params) >= 2 * sum(scenario.pts)``.

    Given the value of ``pts = (M, N, ...)``, it is assumed that
    ``params = [wx1, ..., wxM, x1, ..., xM, wy1, ..., wyN, y1, ..., yN, ...]``.
    Thus, *params* should have ``M`` weights and ``M`` corresponding positions,
    followed by ``N`` weights and ``N`` corresponding positions, with this
    pattern followed for each new dimension of the desired scenario. Any
    remaining parameters will be treated as ``scenario.values``.
"""
    _len = 2 * sum(pts)
    if len(params)  >  _len:  # if Y-values are appended to params
      params, self.values  =  params[:_len], params[_len:]

    self.extend( unflatten(params, pts) )
    return self

  def flatten(self, all=True): #XXX: overwritten.  create standalone instead ?
    """convert a scenario to a single list of parameters

Args:
    all (bool, default=True): if True, append the scenario values

Returns:
    a list of parameters

Notes:
    Given ``scenario.pts = (M, N, ...)``, then the returned list is
    ``params = [wx1, ..., wxM, x1, ..., xM, wy1, ..., wyN, y1, ..., yN, ...]``.
    Thus, *params* will have ``M`` weights and ``M`` corresponding positions,
    followed by ``N`` weights and ``N`` corresponding positions, with this
    pattern followed for each new dimension of the scenario. If *all* is True,
    then the ``scenario.values`` will be appended to the list of parameters.
"""
    params = flatten(self)
    if all: params.extend(self.values) # if Y-values, return those as well
    return params

  # interface
  values = property(__values, __set_values, doc='a list of values corresponding to output data for all point masses in the underlying product measure')
  get_mean_value = mean_value
  pass


#---------------------------------------------
# creators and destructors from parameter list

def _mimic(samples, weights):
  """Generate a product_measure object from a list of N product measure
positions and a list of N weights. The resulting product measure will
mimic the original product measure's statistics, but be larger in size.

For example:
    >>> smp = [[-6,3,6],[-2,4],[1]]
    >>> wts = [[.4,.2,.4],[.5,.5],[1.]]
    >>> c = compose(samples, weights)
    >>> d = _mimic(c.positions, c.weights)
    >>> c[0].center_mass == d[0].center_mass
    True
    >>> c[1].range == d[1].range
    True
    >>> c.npts == d.npts
    False
    >>> c.npts == d[0].npts
    True
"""
  x = list(zip(*samples))               # 'mimic' to a nested list
  w = [weights for i in range(len(x))]  # 'mimic' to a nested list
  return compose(x,w)


def _uniform_weights(samples):
  """generate a nested list of N x 1D weights from a nested list of N x 1D
discrete measure positions, where the weights have norm 1.0 and are uniform.

>>> c.pos
[[1, 2, 3], [4, 5], [6]]
>>> _uniform_weights(c.pos)
[[0.333333333333333, 0.333333333333333, 0.333333333333333], [0.5, 0.5], [1.0]]
"""
  from mystic.math.measures import normalize
  return [normalize([1.]*len(xi), 1.0) for xi in samples]


def _list_of_measures(samples, weights=None):
  """generate a list of N x 1D discrete measures from a nested list of N x 1D
discrete measure positions and a nested list of N x 1D weights.

Note this function does not return a product measure, it returns a list."""
  total = []
  if not weights: weights = _uniform_weights(samples)
  for i in range(len(samples)):
    next = measure()
    for j in range(len(samples[i])):
      next.append(point_mass( samples[i][j], weights[i][j] ))
    total.append(next)
  return total


def compose(samples, weights=None):
  """Generate a product_measure object from a nested list of N x 1D
discrete measure positions and a nested list of N x 1D weights. If weights
are not provided, a uniform distribution with norm = 1.0 will be used."""
  if not weights: weights = _uniform_weights(samples)
  total = _list_of_measures(samples, weights)
  c = product_measure(total)
  return c


def decompose(c):
  """Decomposes a product_measure object into a nested list of
N x 1D discrete measure positions and a nested list of N x 1D weights."""
  from mystic.math.measures import _nested_split
  w, x = _nested_split(flatten(c), c.pts)
  return x, w


#def expand(data, npts):
#  """Generate a scenario object from a dataset. The scenario will have
#uniformly distributed weights and have dimensions given by pts."""
#  positions,values = data.fetch()
#  from mystic.math.measures import _unpack
#  pm = compose( _unpack(positions, npts) )
#  return scenario(pm, values[:pm.npts])


def unflatten(params, npts):
  """Map a list of random variables to N x 1D discrete measures
in a product_measure object."""
  from mystic.math.measures import _nested_split
  w, x = _nested_split(params, npts)
  return compose(x, w)


from itertools import chain #XXX: faster, but sloppy to have as importable
def flatten(c):
  """Flattens a product_measure object into a list."""
  rv = [(i.weights,i.positions) for i in c]
  # now flatten list of lists into just a list
  return list(chain(*chain(*rv))) # faster than mystic.tools.flatten


##### bounds-conserving-mean: borrowed from seismic/seismic.py #####
def bounded_mean(mean_x, samples, xmin, xmax, wts=None):
  from mystic.math.measures import impose_mean, impose_spread
  from mystic.math.measures import spread, mean
  from numpy import asarray
  a = impose_mean(mean_x, samples, wts)
  if min(a) < xmin:   # maintain the bound
    #print("violate lo(a)")
    s = spread(a) - 2*(xmin - min(a)) #XXX: needs compensation (as below) ?
    a = impose_mean(mean_x, impose_spread(s, samples, wts), wts)
  if max(a) > xmax:   # maintain the bound
    #print("violate hi(a)")
    s = spread(a) + 2*(xmax - max(a)) #XXX: needs compensation (as below) ?
    a = impose_mean(mean_x, impose_spread(s, samples, wts), wts)
  return asarray(a)
#####################################################################


#--------------------------------------------------
# constraints solvers and factories for feasibility

# used in self-consistent constraints function c(x) for
#   is_short(x, x') and is_short(x, z)
def norm_wts_constraintsFactory(pts):
  """factory for a constraints function that:
  - normalizes weights
"""
 #from measure import scenario
  def constrain(rv):
    "constrain:  sum(wi)_{k} = 1 for each k in K"
    pm = scenario()
    pm.load(rv, pts)      # here rv is param: w,x,y
    #impose: sum(wi)_{k} = 1 for each k in K
    norm = 1.0
    for i in range(len(pm)):
      w = pm[i].weights
      w[-1] = norm - sum(w[:-1])
      pm[i].weights = w
    rv = pm.flatten(all=True)
    return rv
  return constrain

# used in self-consistent constraints function c(x) for
#   is_short(x, x'), is_short(x, z), and y >= m
def mean_y_norm_wts_constraintsFactory(target, pts):
  """factory for a constraints function that:
  - imposes a mean on scenario values
  - normalizes weights
"""
 #from measure import scenario
  from mystic.math.measures import mean, impose_mean
 #target[0] is target mean
 #target[1] is acceptable deviation
  def constrain(rv):
    "constrain:  y >= m  and  sum(wi)_{k} = 1 for each k in K"
    pm = scenario()
    pm.load(rv, pts)      # here rv is param: w,x,y
    #impose: sum(wi)_{k} = 1 for each k in K
    norm = 1.0
    for i in range(len(pm)):
      w = pm[i].weights
      w[-1] = norm - sum(w[:-1])
      pm[i].weights = w
    #impose: y >= m 
    values, weights = pm.values, pm.weights
    y = float(mean(values, weights))
    if not (y >= float(target[0])):
      pm.values = impose_mean(target[0]+target[1], values, weights)
    rv = pm.flatten(all=True) 
    return rv
  return constrain

def impose_feasible(cutoff, data, guess=None, **kwds):
  """impose shortness on a given scenario

This function attempts to minimize the infeasibility between observed *data*
and a scenario of synthetic data by perforing an optimization on ``w,x,y``
over the given *bounds*.

Args:
    cutoff (float): maximum acceptable deviation from shortness
    data (mystic.math.discrete.scenario): a dataset of observed points
    guess (mystic.math.discrete.scenario, default=None): the synthetic points
    tol (float, default=0.0): maximum acceptable optimizer termination
        for ``sum(infeasibility)``.
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``x' = constraints(x)``,
        where ``x`` is a scenario that has been converted into a list of
        parameters (e.g. with ``scenario.flatten``), and ``x'`` is the list
        of parameters after the encoded constaints have been satisfied.

Returns:
    a scenario with desired shortness

Notes:
    Here, *tol* is used to set the optimization termination for minimizing the
    ``sum(infeasibility)``, while *cutoff* is used in defining the deviation
    from shortness for observed ``x,y`` and synthetic ``x',y'``.

    *guess* can be either a scenario providing initial guess at feasibility,
    or a tuple of the dimensions of the desired scenario, where initial
    values will be chosen at random.
"""
  from numpy import sum, asarray
  from mystic.math.legacydata import dataset
  from mystic.math.distance import lipschitz_distance, infeasibility, _npts
  if guess is None:
    message = "Requires a guess scenario, or a tuple of scenario dimensions."
    raise TypeError(message)
  # get initial guess
  if hasattr(guess, 'pts'): # guess is a scenario
    pts = guess.pts    # number of x
    guess = guess.flatten(all=True)
  else:
    pts = guess        # guess is given as a tuple of 'pts'
    guess = None
  npts = _npts(pts)    # number of Y
  long_form = len(pts) - list(pts).count(2) # can use '2^K compressed format'

  # prepare bounds for solver
  bounds = kwds.pop('bounds', None)
  # if bounds are not set, use the default optimizer bounds
  if bounds is None:
    lower_bounds = []; upper_bounds = []
    for n in pts:  # bounds for n*x in each dimension  (x2 due to weights)
      lower_bounds += [None]*n * 2
      upper_bounds += [None]*n * 2
    # also need bounds for npts*y values
    lower_bounds += [None]*npts
    upper_bounds += [None]*npts
    bounds = lower_bounds, upper_bounds
  bounds = asarray(bounds).T

  # plug in the 'constraints' function:  param' = constraints(param)
  # constraints should impose_mean(y,w), and possibly sum(weights)
  constraints = kwds.pop('constraints', None) # default is no constraints
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  _self = kwds.pop('with_self', True) # default includes self in shortness
  if _self is not False: _self = True
  # tolerance for optimization on sum(y)
  tol = kwds.pop('tol', 0.0) # default
  npop = kwds.pop('npop', 20) #XXX: tune npop?
  maxiter = kwds.pop('maxiter', 1000) #XXX: tune maxiter?

  # if no guess was made, then use bounds constraints
  if guess is None:
    if npop:
      guess = bounds
    else:  # fmin_powell needs a list params (not bounds)
      guess = [(a + b)/2. for (a,b) in bounds]

  # construct cost function to reduce sum(lipschitz_distance)
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where:  cost = | sum(lipschitz_distance) | """
    _data = dataset()
    _pm = scenario()
    _pm.load(rv, pts)      # here rv is param: w,x,y
    if not long_form:
      positions = _pm.select(*range(npts))
    else: positions = _pm.positions
    _data.load( data.coords, data.values )                   # LOAD static
    if _self:
      _data.load( positions, _pm.values )                    # LOAD dynamic
    _data.lipschitz = data.lipschitz                         # LOAD L
    Rv = lipschitz_distance(_data.lipschitz, _pm, _data, tol=cutoff, **kwds)
    v = infeasibility(Rv, cutoff)
    return abs(sum(v))

  # construct and configure optimizer
  debug = False  #!!!
  maxfun = 1e+6
  crossover = 0.9; percent_change = 0.9
  ftol = abs(tol); gtol = None

  if debug:
    print("lower bounds: %s" % bounds.T[0])
    print("upper bounds: %s" % bounds.T[1])
  # print("initial value: %s" % guess)
  # use optimization to get feasible points
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor
  from mystic.strategy import Best1Bin, Best1Exp
  evalmon = Monitor();  stepmon = Monitor(); strategy = Best1Exp
  if debug: stepmon = VerboseMonitor(10)  #!!!
  if npop: # use VTR
    results = diffev2(cost, guess, npop, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      cross=crossover, scale=percent_change, strategy=strategy,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  else: # use VTR
    results = fmin_powell(cost, guess, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  # repack the results
  pm = scenario()
  pm.load(results[0], pts)            # params: w,x,y
 #if debug: print("final cost: %s" % results[1])
  if debug and results[2] >= maxiter: # iterations
    print("Warning: constraints solver terminated at maximum iterations")
 #func_evals = results[3]           # evaluation
  return pm


def impose_valid(cutoff, model, guess=None, **kwds):
  """impose model validity on a given scenario

This function attempts to minimize the graph distance between reality (data),
``y = G(x)``, and an approximating function, ``y' = F(x')``, by perforing an
optimization on ``w,x,y`` over the given *bounds*.

Args:
    cutoff (float): maximum acceptable model invalidity ``|y - F(x')|``.
    model (func): the model function, ``y' = F(x')``.
    guess (scenario, default=None): a scenario, defines ``y = G(x)``.
    hausdorff (bool, default=False): hausdorff ``norm``, where if given,
        then ``ytol = |y - F(x')| + |x - x'|/norm``
    xtol (float, default=0.0): maximum acceptable pointwise graphical distance
        between model and reality.
    tol (float, default=0.0): maximum acceptable optimizer termination
        for ``sum(graphical distances)``.
    bounds (tuple, default=None): ``(all lower bounds, all upper bounds)``
    constraints (func, default=None): a function ``x' = constraints(x)``,
        where ``x`` is a scenario that has been converted into a list of
        parameters (e.g. with ``scenario.flatten``), and ``x'`` is the list
        of parameters after the encoded constaints have been satisfied.

Returns:
    a scenario with the desired model validity

Notes:
    *xtol* defines the n-dimensional base of a pilar of height *cutoff*,
    centered at each point. The region inside the pilar defines the space
    where a "valid" model must intersect. If *xtol* is not specified, then
    the base of the pilar will be a dirac at ``x' = x``. This function
    performs an optimization to find a set of points where the model is valid.
    Here, *tol* is used to set the optimization termination for minimizing the
    ``sum(graphical_distances)``, while *cutoff* is used in defining the
    graphical distance between ``x,y`` and ``x',F(x')``.

    *guess* can be either a scenario providing initial guess at validity,
    or a tuple of the dimensions of the desired scenario, where initial
    values will be chosen at random.
"""
  #FIXME: there are a lot of undocumented kwds (see below)
  from numpy import sum as _sum, asarray
  from mystic.math.distance import graphical_distance, infeasibility, _npts
  if guess is None:
    message = "Requires a guess scenario, or a tuple of scenario dimensions."
    raise TypeError(message)
  # get initial guess
  if hasattr(guess, 'pts'): # guess is a scenario
    pts = guess.pts    # number of x
    guess = guess.flatten(all=True)
  else:
    pts = guess        # guess is given as a tuple of 'pts'
    guess = None
  npts = _npts(pts)    # number of Y

  # prepare bounds for solver
  bounds = kwds.pop('bounds', None)
  # if bounds are not set, use the default optimizer bounds
  if bounds is None:
    lower_bounds = []; upper_bounds = []
    for n in pts:  # bounds for n*x in each dimension  (x2 due to weights)
      lower_bounds += [None]*n * 2
      upper_bounds += [None]*n * 2
    # also need bounds for npts*y values
    lower_bounds += [None]*npts
    upper_bounds += [None]*npts
    bounds = lower_bounds, upper_bounds
  bounds = asarray(bounds).T

  # plug in the 'constraints' function:  param' = constraints(param)
  constraints = kwds.pop('constraints', None) # default is no constraints
  if not constraints:  # if None (default), there are no constraints
    constraints = lambda x: x

  # 'wiggle room' tolerances
  ipop = kwds.pop('ipop', 10) #XXX: tune ipop (inner optimization)?
  imax = kwds.pop('imax', 10) #XXX: tune imax (inner optimization)?
  # tolerance for optimization on sum(y)
  tol = kwds.pop('tol', 0.0) # default
  npop = kwds.pop('npop', 20) #XXX: tune npop (outer optimization)?
  maxiter = kwds.pop('maxiter', 1000) #XXX: tune maxiter (outer optimization)?

  # if no guess was made, then use bounds constraints
  if guess is None:
    if npop:
      guess = bounds
    else:  # fmin_powell needs a list params (not bounds)
      guess = [(a + b)/2. for (a,b) in bounds]

  # construct cost function to reduce sum(infeasibility)
  def cost(rv):
    """compute cost from a 1-d array of model parameters,
    where: cost = | sum( infeasibility ) | """
    # converting rv to scenario
    points = scenario()
    points.load(rv, pts)
    # calculate infeasibility
    Rv = graphical_distance(model, points, ytol=cutoff, ipop=ipop, \
                                                        imax=imax, **kwds)
    v = infeasibility(Rv, cutoff)
    # converting v to E
    return _sum(v) #XXX: abs ?

  # construct and configure optimizer
  debug = False  #!!!
  maxfun = 1e+6
  crossover = 0.9; percent_change = 0.8
  ftol = abs(tol); gtol = None #XXX: optimally, should be VTRCOG...

  if debug:
    print("lower bounds: %s" % bounds.T[0])
    print("upper bounds: %s" % bounds.T[1])
  # print("initial value: %s" % guess)
  # use optimization to get model-valid points
  from mystic.solvers import diffev2, fmin_powell
  from mystic.monitors import Monitor, VerboseMonitor
  from mystic.strategy import Best1Bin, Best1Exp
  evalmon = Monitor();  stepmon = Monitor(); strategy = Best1Exp
  if debug: stepmon = VerboseMonitor(2)  #!!!
  if npop: # use VTR
    results = diffev2(cost, guess, npop, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      cross=crossover, scale=percent_change, strategy=strategy,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  else: # use VTR
    results = fmin_powell(cost, guess, ftol=ftol, gtol=gtol, bounds=bounds,\
                      maxiter=maxiter, maxfun=maxfun, constraints=constraints,\
                      evalmon=evalmon, itermon=stepmon,\
                      full_output=1, disp=0, handler=False)
  # repack the results
  pm = scenario()
  pm.load(results[0], pts)            # params: w,x,y
 #if debug: print("final cost: %s" % results[1])
  if debug and results[2] >= maxiter: # iterations
    print("Warning: constraints solver terminated at maximum iterations")
 #func_evals = results[3]           # evaluation
  return pm


# backward compatibility
point = point_mass
dirac_measure = measure


if __name__ == '__main__':
  from mystic.math.distance import *
  model = lambda x:sum(x)
  a = [0,1,9,8, 1,0,4,6, 1,0,1,2, 0,1,2,3,4,5,6,7]
  feasability = 0.0; deviation = 0.01
  validity = 5.0; wiggle = 1.0
  y_mean = 5.0; y_buffer = 0.0
  L = [.75,.5,.25]
  bc = [(0,7,2),(3,0,2),(2,0,3),(1,0,3),(2,4,2)]
  bv = [5,3,1,4,8]
  pts = (2,2,2)
  from mystic.math.legacydata import dataset
  data = dataset()
  data.load(bc, bv)
  data.lipschitz = L
  pm = scenario()
  pm.load(a, pts)
  pc = pm.positions
  pv = pm.values
  #---
  _data = dataset()
  _data.load(bc, bv)
  _data.load(pc, pv)
  _data.lipschitz = data.lipschitz
  from numpy import sum
  ans = sum(lipschitz_distance(L, pm, _data))
  print("original: %s @ %s\n" % (ans, a))
 #print("pm: %s" % pm)
 #print("data: %s" % data)
  #---
  lb = [0,.5,-100,-100,  0,.5,-100,-100,  0,.5,-100,-100,   0,0,0,0,0,0,0,0]
  ub = [.5,1, 100, 100,  .5,1, 100, 100,  .5,1, 100, 100,   9,9,9,9,9,9,9,9]
  bounds = (lb,ub)

  _constrain = mean_y_norm_wts_constraintsFactory((y_mean,y_buffer), pts)
  results = impose_feasible(feasability, data, guess=pts, tol=deviation, \
                            bounds=bounds, constraints=_constrain)
  from mystic.math.measures import mean
  print("solved: %s" % results.flatten(all=True))
  print("mean(y): %s >= %s" % (mean(results.values, results.weights), y_mean))
  print("sum(wi): %s == 1.0" % [sum(w) for w in results.wts])

  print("\n---------------------------------------------------\n")

  bc = bc[:-2]
  ids = ['1','2','3']
  t = dataset()
  t.load(bc, list(map(model, bc)), ids)
  t.update(t.coords, list(map(model, t.coords)))
# r = dataset()
# r.load(t.coords, t.values)
# L = [0.1, 0.0, 0.0]
  print("%s" % t)
  print("L: %s" % L)
  print("shortness:")
  print(lipschitz_distance(L, t, t, tol=0.0))
  
  print("\n---------------------------------------------------\n")

  print("Y: %s" % str(results.values))
  print("sum(wi): %s == 1.0" % [sum(w) for w in results.wts])


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

""" NOTE:
rosen rewritten as a pickleable function
"""

def rosen(coeffs):
    """evaluates n-dimensional Rosenbrock function for a list of coeffs

minimum is f(x)=0.0 at xi=1.0"""
    from numpy import sum as numpysum
    from numpy import asarray

    #ensure that there are 2 coefficients
    x = [1]*2
    x[:len(coeffs)]=coeffs
    x = asarray(x) #XXX: must be a numpy.array
    return numpysum(100.0*(x[1:]-x[:-1]**2.0)**2.0 + (1-x[:-1])**2.0)


if __name__=='__main__':
    target = [1., 1., 1.]
    print(target)
    print("")
    print(rosen(target))



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

import os
import sys
# drop support for older python
unsupported = None
if sys.version_info < (2, 7):
    unsupported = 'Versions of Python before 2.7 are not supported'
elif (3, 0) <= sys.version_info < (3, 7):
    unsupported = 'Versions of Python before 3.7 are not supported'
if unsupported:
    raise ValueError(unsupported)

# get distribution meta info
here = os.path.abspath(os.path.dirname(__file__))
meta_fh = open(os.path.join(here, 'mystic/__init__.py'))
try:
    meta = {}
    for line in meta_fh:
        if line.startswith('__version__'):
            VERSION = line.split()[-1].strip("'").strip('"')
            break
    meta['VERSION'] = VERSION
    for line in meta_fh:
        if line.startswith('__author__'):
            AUTHOR = line.split(' = ')[-1].strip().strip("'").strip('"')
            break
    meta['AUTHOR'] = AUTHOR
    LONG_DOC = ""
    DOC_STOP = "FAKE_STOP_12345"
    for line in meta_fh:
        if LONG_DOC:
            if line.startswith(DOC_STOP):
                LONG_DOC = LONG_DOC.strip().strip("'").strip('"').lstrip()
                break
            else:
                LONG_DOC += line
        elif line.startswith('__doc__'):
            DOC_STOP = line.split(' = ')[-1]
            LONG_DOC = "\n"
    meta['LONG_DOC'] = LONG_DOC
finally:
    meta_fh.close()

# get version numbers, long_description, etc
AUTHOR = meta['AUTHOR']
VERSION = meta['VERSION']
LONG_DOC = meta['LONG_DOC'] #FIXME: near-duplicate of README.md
#LICENSE = meta['LICENSE'] #FIXME: duplicate of LICENSE
AUTHOR_EMAIL = 'mmckerns@uqfoundation.org'

# check if setuptools is available
try:
    from setuptools import setup
    from setuptools.dist import Distribution
    has_setuptools = True
except ImportError:
    from distutils.core import setup
    Distribution = object
    has_setuptools = False

# build the 'setup' call
setup_kwds = dict(
    name='mystic',
    version=VERSION,
    description='highly-constrained non-convex optimization and uncertainty quantification',
    long_description = LONG_DOC,
    author = AUTHOR,
    author_email = AUTHOR_EMAIL,
    maintainer = AUTHOR,
    maintainer_email = AUTHOR_EMAIL,
    license = '3-clause BSD',
    platforms = ['Linux', 'Windows', 'Mac'],
    url = 'https://github.com/uqfoundation/mystic',
    download_url = 'https://pypi.org/project/mystic/#files',
    project_urls = {
        'Documentation':'http://mystic.rtfd.io',
        'Source Code':'https://github.com/uqfoundation/mystic',
        'Bug Tracker':'https://github.com/uqfoundation/mystic/issues',
    },
    python_requires = '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
    packages = ['mystic','mystic.models','mystic.math','mystic.cache',
                'mystic.tests'],
    package_dir = {'mystic':'mystic','mystic.models':'models',
                   'mystic.math':'mystic/math','mystic.cache':'cache',
                   'mystic.tests':'tests'},
    scripts=['scripts/mystic_log_reader',
             'scripts/mystic_model_plotter',
             'scripts/mystic_collapse_plotter',
             'scripts/support_convergence',
             'scripts/support_hypercube',
             'scripts/support_hypercube_measures',
             'scripts/support_hypercube_scenario'],
)

# force python-, abi-, and platform-specific naming of bdist_wheel
class BinaryDistribution(Distribution):
    """Distribution which forces a binary package with platform name"""
    def has_ext_modules(foo):
        return True

# define dependencies
sysversion = sys.version_info[:2]
dill_version = 'dill>=0.3.5.1'
klepto_version = 'klepto>=0.2.2'
pathos_version = 'pathos>=0.2.9'
pyina_version = 'pyina>=0.2.6'
cython_version = 'cython>=0.29.22' #XXX: required to build numpy from source
try:
    import ctypes # if using `pypy`, pythonapi is not found
    IS_PYPY = not hasattr(ctypes, 'pythonapi')
    IS_PYPY2 = IS_PYPY and sysversion < (3,0)
except:
    IS_PYPY = False
    IS_PYPY2 = False
if sysversion < (2,6) or sysversion == (3,0) or sysversion == (3,1):
    numpy_version = 'numpy>=1.0, <1.8.0'
    sympy_version = 'sympy>=0.6.7, <1.1'
    scipy_version = 'scipy>=0.6.0, <0.17.0'
    mpmath_version = 'mpmath>=0.19, <1.0.0'
    matplotlib_version = 'matplotlib>=0.91, <2.0.0'
elif sysversion == (2,6) or sysversion == (3,2) or sysversion == (3,3):
    numpy_version = 'numpy>=1.0, <1.12.0'
    sympy_version = 'sympy>=0.6.7, <1.1'
    scipy_version = 'scipy>=0.6.0, <1.0.0'
    mpmath_version = 'mpmath>=0.19, <1.0.0'
    matplotlib_version = 'matplotlib>=0.91, <2.0.0'
elif IS_PYPY2:
    numpy_version = 'numpy>=1.0, <1.16.0'
    sympy_version = 'sympy>=0.6.7, <1.1'
    scipy_version = 'scipy>=0.6.0, <1.3.0'
    mpmath_version = 'mpmath>=0.19, <1.2.1' #XXX: != 1.2.1
    matplotlib_version = 'matplotlib>=0.91, <3.0.0'
elif sysversion == (2,7) or sysversion == (3,4):
    numpy_version = 'numpy>=1.0, <1.17.0'
    sympy_version = 'sympy>=0.6.7, <1.1'
    scipy_version = 'scipy>=0.6.0, <1.3.0'
    mpmath_version = 'mpmath>=0.19'
    matplotlib_version = 'matplotlib>=0.91, <3.0.0'
elif sysversion == (3,5):
    numpy_version = 'numpy>=1.0, <1.19.0'
    sympy_version = 'sympy>=0.6.7, <1.7'
    scipy_version = 'scipy>=0.6.0, <1.5.0'
    mpmath_version = 'mpmath>=0.19'
    matplotlib_version = 'matplotlib>=0.91, <3.1.0'
elif sysversion == (3,6):# or IS_PYPY
    numpy_version = 'numpy>=1.0, <1.20.0'
    sympy_version = 'sympy>=0.6.7'#, <0.7.4'
    scipy_version = 'scipy>=0.6.0, <1.6.0'
    mpmath_version = 'mpmath>=0.19'
    matplotlib_version = 'matplotlib>=0.91, <3.4.0'
else:
    numpy_version = 'numpy>=1.0'
    sympy_version = 'sympy>=0.6.7'#, <0.7.4'
    scipy_version = 'scipy>=0.6.0'
    mpmath_version = 'mpmath>=0.19'
    matplotlib_version = 'matplotlib>=0.91' #XXX: kiwisolver-1.3.0
# add dependencies
depend = [dill_version, klepto_version, numpy_version, sympy_version, mpmath_version]
extras = {'math': [scipy_version], 'parallel': [pathos_version, pyina_version], 'plotting': [matplotlib_version]}
# update setup kwds
if has_setuptools:
    setup_kwds.update(
        zip_safe=False,
        # distclass=BinaryDistribution,
        install_requires=depend,
        extras_require=extras,
    )

# call setup
setup(**setup_kwds)

# if dependencies are missing, print a warning
try:
    import numpy
    import sympy
    import mpmath
    import dill
    import klepto
    #import cython
    #import scipy
    #import matplotlib #XXX: has issues being zip_safe
    #import pathos
    #import pyina
except ImportError:
    print("\n***********************************************************")
    print("WARNING: One of the following dependencies is unresolved:")
    print("    %s" % numpy_version)
    print("    %s" % sympy_version)
    print("    %s" % mpmath_version)
    print("    %s" % dill_version)
    print("    %s" % klepto_version)
    #print("    %s" % cython_version)
    print("    %s (optional)" % scipy_version)
    print("    %s (optional)" % matplotlib_version)
    print("    %s (optional)" % pathos_version)
    print("    %s (optional)" % pyina_version)
    print("***********************************************************\n")


if __name__=='__main__':
    pass

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Minimize Rosenbrock's Function with Nelder-Mead.
    - Plot of Rosenbrock's function minimum.

Demonstrates:
    - standard models
    - minimal solver interface
"""

# Nelder-Mead solver
from mystic.solvers import fmin

# Rosenbrock function
from mystic.models import rosen

# tools
import matplotlib.pyplot as plt


if __name__ == '__main__':

    print("Nelder-Mead Simplex")
    print("===================")

    # initial guess
    x0 = [0.8,1.2,0.7]

    # use Nelder-Mead to minimize the Rosenbrock function
    solution = fmin(rosen,x0)
    print(solution)
 
    # plot the Rosenbrock function (one plot per axis)
    x = [0.01*i for i in range(200)]
    plt.plot(x,[rosen([i,1.,1.]) for i in x])
    plt.plot(x,[rosen([1.,i,1.]) for i in x])
    plt.plot(x,[rosen([1.,1.,i]) for i in x])

    # plot the solved minimum (for x)
    plt.plot([solution[0]],[rosen(solution)],'bo')

    # draw the plot
    plt.title("minimium of Rosenbrock's function")
    plt.xlabel("x, y, z")
    plt.ylabel("f(i) = Rosenbrock's function")
    plt.legend(["f(x,1,1)","f(1,y,1)","f(1,1,z)"])
    plt.show()

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.models import rosen
from mystic.solvers import *
from mystic.termination import VTRChangeOverGeneration
from mystic.monitors import VerboseMonitor, Monitor
from mystic.tools import random_seed
random_seed(123)
lb, ub = [-100.]*3, [100]*3
interval = None

if interval:
  _stepmon = VerboseMonitor(interval)
else:
  _stepmon = Monitor()
_term = VTRChangeOverGeneration(generations=200)
_solver = DifferentialEvolutionSolver(3, 20)#40)
_solver.SetRandomInitialPoints(lb,ub)
_solver.SetStrictRanges(lb,ub)
_solver.SetTermination(_term)
_solver.SetGenerationMonitor(_stepmon)
_solver.SetEvaluationLimits(100, 1000)
_solver.Solve(rosen)

_energy = _solver.bestEnergy
_solution =  _solver.bestSolution
_population = _solver.population

_solver.SetEvaluationLimits(10000, 100000)
_solver.Solve()

_energy = _solver.bestEnergy
_solution = _solver.bestSolution
_population = _solver.population

# again, with same method order
random_seed(123)
if interval:
  stepmon = VerboseMonitor(interval)
else:
  stepmon = Monitor()
term = VTRChangeOverGeneration(generations=200)
solver = DifferentialEvolutionSolver(3, 20)#40)
solver.SetRandomInitialPoints(lb,ub)
solver.SetStrictRanges(lb,ub)
solver.SetTermination(term)
solver.SetGenerationMonitor(stepmon)
solver.SetEvaluationLimits(100, 1000)
solver.Solve(rosen)

energy = solver.bestEnergy
solution =  solver.bestSolution
population = solver.population

solver.SetEvaluationLimits(10000, 100000)
solver.Solve()

energy = solver.bestEnergy
solution = solver.bestSolution
population = solver.population

# compare results
assert energy == _energy
assert all(solution == _solution)

# again, with different method order
random_seed(123)
if interval:
  stepmon = VerboseMonitor(interval)
else:
  stepmon = Monitor()
term = VTRChangeOverGeneration(generations=200)
solver = DifferentialEvolutionSolver(3, 20)#40)
solver.SetRandomInitialPoints(lb,ub)
solver.SetGenerationMonitor(stepmon)
solver.SetEvaluationLimits(100, 1000)
solver.SetTermination(term)
solver.SetStrictRanges(lb,ub)
solver.Solve(rosen)

energy = solver.bestEnergy
solution =  solver.bestSolution
population = solver.population

solver.SetEvaluationLimits(10000, 100000)
solver.Solve()

energy = solver.bestEnergy
solution = solver.bestSolution
population = solver.population

# compare results
assert energy == _energy
assert all(solution == _solution)

# again, but focused on method order for population
random_seed(123)
if interval:
  stepmon = VerboseMonitor(interval)
else:
  stepmon = Monitor()
term = VTRChangeOverGeneration(generations=200)
solver = DifferentialEvolutionSolver(3, 20)#40)
solver.SetGenerationMonitor(stepmon)
solver.SetEvaluationLimits(100, 1000)
solver.SetTermination(term)
solver.SetStrictRanges(lb,ub)
solver.SetRandomInitialPoints(lb,ub)
solver.Solve(rosen)

energy = solver.bestEnergy
solution =  solver.bestSolution
population = solver.population

solver.SetEvaluationLimits(10000, 100000)
solver.Solve()

energy = solver.bestEnergy
solution = solver.bestSolution
population = solver.population

# compare results
assert energy == _energy
assert all(solution == _solution)

# start over... this time focus on methods called on the restart
interval = None
random_seed(213)
if interval:
  _stepmon2 = VerboseMonitor(interval)
else:
  _stepmon2 = Monitor()
_term2 = VTRChangeOverGeneration(generations=2000)
_solver2 = DifferentialEvolutionSolver(3, 20)#40)
_solver2.SetEvaluationLimits(100, 1000)
_solver2.SetTermination(_term2)
_solver2.SetRandomInitialPoints(lb,ub)
_solver2.Solve(rosen)

_energy2 = _solver2.bestEnergy
_solution2 =  _solver2.bestSolution
_population2 = _solver2.population

_term2 = VTRChangeOverGeneration(generations=200)
_solver2.SetStrictRanges(lb,ub)
_solver2.SetEvaluationLimits(new=True)
_solver2.SetGenerationMonitor(_stepmon2)
_solver2.SetTermination(_term2)
_solver2.Solve()

_energy2 = _solver2.bestEnergy
_solution2 = _solver2.bestSolution
_population2 = _solver2.population

# again, but swap method order for restart
random_seed(213)
if interval:
  stepmon2 = VerboseMonitor(interval)
else:
  stepmon2 = Monitor()
term2 = VTRChangeOverGeneration(generations=2000)
solver2 = DifferentialEvolutionSolver(3, 20)#40)
solver2.SetEvaluationLimits(100, 1000)
solver2.SetTermination(term2)
solver2.SetRandomInitialPoints(lb,ub)
solver2.Solve(rosen)

energy2 = solver2.bestEnergy
solution2 =  solver2.bestSolution
population2 = solver2.population

term2 = VTRChangeOverGeneration(generations=200)
solver2.SetTermination(term2)
solver2.SetGenerationMonitor(stepmon2)
solver2.SetEvaluationLimits(new=True)
solver2.SetStrictRanges(lb,ub)
solver2.Solve()

energy2 = solver2.bestEnergy
solution2 = solver2.bestSolution
population2 = solver2.population

# compare results
assert energy2 == _energy2
assert all(solution2 == _solution2)

# start over... and change so initialize population in the restart
interval = None
random_seed(123)
if interval:
  _stepmon3 = VerboseMonitor(interval)
else:
  _stepmon3 = Monitor()
_term3 = VTRChangeOverGeneration(generations=2000)
_solver3 = DifferentialEvolutionSolver(3, 20)#40)
_solver3.SetEvaluationLimits(100, 1000)
_solver3.SetTermination(_term3)
_solver3.SetGenerationMonitor(_stepmon3)
_solver3.SetRandomInitialPoints(lb,ub)
_solver3.Solve(rosen)

_energy3 = _solver3.bestEnergy
_solution3 =  _solver3.bestSolution
_population3 = _solver3.population

_lb, _ub = [-10.]*3, [10]*3
_term3 = VTRChangeOverGeneration(generations=200)
_solver3.SetRandomInitialPoints(_lb,_ub) #FIXME: pretty much causes a flatline
_solver3.SetStrictRanges(lb,ub)          #       regardless of _lb,_ub value
_solver3.SetEvaluationLimits(200, 2000, new=True)# check if pop is uniform
_solver3.SetTermination(_term3)
_solver3.Solve()

_energy3 = _solver3.bestEnergy
_solution3 = _solver3.bestSolution
_population3 = _solver3.population
# FIXME: when population becomes uniform, the solver will get stuck
#        would be good to be able to inject some randomness into a restart

# again, but swap method order for restart
random_seed(123)
if interval:
  stepmon3 = VerboseMonitor(interval)
else:
  stepmon3 = Monitor()
term3 = VTRChangeOverGeneration(generations=2000)
solver3 = DifferentialEvolutionSolver(3, 20)#40)
solver3.SetEvaluationLimits(100, 1000)
solver3.SetTermination(term3)
solver3.SetGenerationMonitor(stepmon3)
solver3.SetRandomInitialPoints(lb,ub)
solver3.Solve(rosen)

energy3 = solver3.bestEnergy
solution3 =  solver3.bestSolution
population3 = solver3.population

term3 = VTRChangeOverGeneration(generations=200)
solver3.SetTermination(term3)
solver3.SetEvaluationLimits(200, 2000, new=True)
solver3.SetStrictRanges(lb,ub)
solver3.SetRandomInitialPoints(_lb,_ub)
solver3.Solve()

energy3 = solver3.bestEnergy
solution3 = solver3.bestSolution
population3 = solver3.population

# compare results
assert energy3 == _energy3
assert all(solution3 == _solution3)

# TODO: start over... but work with Step instead of Solve

# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Minimize Rosenbrock's Function with Powell's method.

Demonstrates:
    - standard models
    - minimal solver interface
    - parameter constraints solver and constraints factory decorator
    - statistical parameter constraints
    - customized monitors
"""

# Powell's Directonal solver
from mystic.solvers import fmin_powell

# Rosenbrock function
from mystic.models import rosen

# tools
from mystic.monitors import VerboseMonitor
from mystic.math.measures import mean, impose_mean
from mystic.math import almostEqual


if __name__ == '__main__':

    print("Powell's Method")
    print("===============")

    # initial guess
    x0 = [0.8,1.2,0.7]

    # use the mean constraints factory decorator
    from mystic.constraints import with_mean

    # define constraints function
    @with_mean(1.0)
    def constraints(x):
        # constrain the last x_i to be the same value as the first x_i
        x[-1] = x[0]
        return x

    # configure monitor
    stepmon = VerboseMonitor(1)

    # use Powell's method to minimize the Rosenbrock function
    solution = fmin_powell(rosen,x0,constraints=constraints,itermon=stepmon)
    print(solution)
 
# end of file



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from g01 import objective, bounds, xs, ys
from g01_alt import penalty1, penalty2, penalty3, penalty4, penalty5, \
                    penalty6, penalty7, penalty8, penalty9

from mystic.constraints import as_constraint
from mystic.penalty import linear_inequality
from mystic.coupler import and_ as combined

penalties = (penalty1,penalty2,penalty3,penalty4,penalty5,\
             penalty6,penalty7,penalty8,penalty9)

penalty = combined(*[linear_inequality(pi)(lambda x:0.) for pi in penalties])
solver = as_constraint(penalty)



if __name__ == '__main__':
    x = [0]*len(xs)

    from mystic.solvers import fmin_powell, diffev
    from mystic.math import almostEqual

    result = fmin_powell(objective, x0=x, bounds=bounds, penalty=penalty, disp=False, full_output=True)

    assert almostEqual(result[0], xs, tol=1e-2)
    assert almostEqual(result[1], ys, tol=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Sets up Storn and Price's Polynomial 'Fitting' Problem for ChebyshevT16

Reference:

[1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
Heuristic for Global Optimization over Continuous Spaces. Journal of Global
Optimization 11: 341-359, 1997.
"""

from mystic.solvers import DifferentialEvolutionSolver
from mystic.termination import ChangeOverGeneration, VTR
from mystic.strategy import Best1Exp, Best1Bin, Rand1Exp, Best2Bin, Best2Exp
from mystic.math import poly1d
from mystic.monitors import VerboseMonitor
from mystic.models.poly import chebyshev16cost

from mystic.tools import random_seed
random_seed(123)

# get the target coefficients
from mystic.models.poly import chebyshev16coeffs as Chebyshev16

def ChebyshevCost(trial):
    """
The costfunction for the fitting problem.
70 evaluation points between [-1, 1] with two end points
    """
    return chebyshev16cost(trial,M=70)*100


ND = 17
NP = 100
MAX_GENERATIONS = 5000

from test_ffit import plot_solution

def main():
    solver = DifferentialEvolutionSolver(ND, NP)
    solver.SetRandomInitialPoints()
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(VerboseMonitor(10))
  
    #strategy = Best1Exp
    #strategy = Best1Bin
    #strategy = Best2Bin
    strategy = Best2Exp

    solver.Solve(ChebyshevCost, termination=VTR(0.0001), strategy=strategy, \
                 CrossProbability=1.0, ScalingFactor=0.6)

    solution = solver.Solution()
  
    print("\nsolved: ")
    print(poly1d(solution))
    print("\ntarget: ")
    print(poly1d(Chebyshev16))
   #print("actual coefficients vs computed:")
   #for actual,computed in zip(Chebyshev16, solution):
   #    print("%f %f" % (actual, computed))

    plot_solution(solution, Chebyshev16)


if __name__ == '__main__':
    from timeit import Timer
    t = Timer("main()", "from __main__ import main")
    timetaken =  t.timeit(number=1)
    print("\nCPU Time: %s" % timetaken)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = _doc = """
This is part of Hock and Schittkowski's test suite in [1], with function
definitions drawn from [1] and [2].

References:
    1. Hock, W. and Schittkowski, K. "Test Examples for Nonlinear Programming
       Codes", Lecture Notes in Economics and Mathematical Systems, Vol. 187,
       Springer, 1981. http://www.ai7.uni-bayreuth.de/test_problem_coll.pdf
    2. Paviani, D.A. "A new method for the solution of the general
       nonlinear programming problem", Ph.D. dissertation, The University
       of Texas, Austin, TX, 1969.
"""
from .abstract_model import AbstractFunction

from math import sin, cos, sqrt, pi, exp, log
from numpy import inf

class Paviani(AbstractFunction):
    __doc__ = \
    """a Paviani's function generator

Paviani's function [1,2] is a relatively flat basin that
quickly jumps to infinity for x_i >= 10 or x_i <= 2. The
global minimum is located near the corner of one of the
basin corners. There are local minima in the corners
ajacent to the global minima.

The generated function f(x) is identical to function (110)
of [1], where len(x) >= 0.
    """ + _doc
    def __init__(self, ndim=10): # is n-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Paviani's function for a list of coeffs

f(x) = f_0(x) - f_1(x)

Where:
f_0(x) = \sum_(i=0)^(N-1) (\ln(x_i - 2)^2 + \ln(10 - x_i)^2)
and
f_1(x) = \prod_(i=0)^(N-1) x_(i)^(.2)

Inspect with mystic_model_plotter using::
    mystic.models.paviani -b "2:10:.1, 2:10:.1" -d

For N=1, the minimum is f(x)=2.133838 at x_i=8.501586,
for N=3, the minimum is f(x)=7.386004 at x_i=8.589578,
for N=5, the minimum is f(x)=9.730525 at x_i=8.740743,
for N=8, the minimum is f(x)=-3.411859 at x_i=9.086900,
for N=10, the minimum is f(x)=-45.778470 at x_i=9.350241."""
        s = 0.
        for x in coeffs:
            if x >= 10. or x <= 2.:
                s = inf
                break
            s += log(x-2.)**2 + log(10.-x)**2
        p = reduce(lambda x,y:x*y, coeffs)**(.2)
        return s - p

    minimizers = None #XXX: there are also some local minima
    pass


# cleanup
del _doc

# prepared instances
paviani = Paviani().function


# End of file



#!/usr/bin/env python
# ******NOTICE***************
# optimize.py module by Travis E. Oliphant
#
# You may copy and use this module as you see fit with no
# guarantee implied provided you keep this notice in all copies.
# *****END NOTICE************
#
# Forked by: Mike McKerns (February 2009)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2009-2016 California Institute of Technology.
# Copyright (c) 2016-2022 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
""" local copy of scipy.optimize.linesearch """

def line_search(f, myfprime, xk, pk, gfk, old_fval, old_old_fval,
                args=(), c1=1e-4, c2=0.9, amax=50):

    try: #XXX: break dependency on scipy.optimize.linesearch
        from scipy.optimize import linesearch
        alpha_k, fc, gc, old_fval, old_old_fval, gfkp1 = \
                         linesearch.line_search(f,myfprime,xk,pk,gfk,\
                         old_fval,old_old_fval,args,c1,c2,amax)
    except ImportError:
        alpha_k = None
        fc = 0
        gc = 0
        gfkp1 = gfk #XXX: or None ?

    return alpha_k, fc, gc, old_fval, old_old_fval, gfkp1

